#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import inspect
import logging
import textwrap
import typing
import warnings
from abc import ABC
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

import pandas as pd
from argilla._constants import OPENAI_SEPARATOR, OPENAI_WHITESPACE
from argilla.client.feedback.schemas import (
    FeedbackRecord,
    LabelQuestion,
    MultiLabelQuestion,
    RankingQuestion,
    RatingQuestion,
    TextField,
    TextQuestion,
)
from argilla.client.feedback.training.schemas.defaults import (
    QuestionAnsweringDefaults,
    SentenceSimilarityDefaults,
    TextClassificationDefaults,
)
from argilla.client.feedback.training.schemas.return_types import (
    ChatCompletionReturnTypes,
    DPOReturnTypes,
    PPOReturnTypes,
    QuestionAnsweringReturnTypes,
    RMReturnTypes,
    SentenceSimilarityReturnTypes,
    SFTReturnTypes,
    TextClassificationReturnTypes,
)
from argilla.client.feedback.unification import (
    LabelQuestionUnification,
    MultiLabelQuestionUnification,
    RankingQuestionUnification,
    RatingQuestionUnification,
)
from argilla.client.models import Framework
from argilla.pydantic_v1 import BaseModel
from argilla.utils.dependency import require_dependencies, requires_dependencies

_LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import datasets
    import spacy
    from argilla.client.feedback.dataset import FeedbackDataset
    from sentence_transformers import InputExample


class TrainingData(ABC):
    formatting_func: Optional[BaseModel] = None
    defaults: Optional[BaseModel] = None
    _formatting_func_return_types = None
    _supported_frameworks_names: list = []

    @property
    def formatting_func_return_types(self) -> BaseModel:
        return self._formatting_func_return_types

    @property
    def supported_frameworks(self) -> List[Framework]:
        return [Framework(name) for name in self._supported_frameworks_names]

    def __repr__(self) -> str:
        def get_func_repr(func: Callable) -> str:
            func_source = inspect.getsource(func)
            return func_source

        return textwrap.dedent(
            f"{self.__class__.__name__}\nformatting_func:\n{get_func_repr(self.formatting_func)}"
            if self.formatting_func
            else f"\ndefaults:\n{self.defaults})"
        )

    def test_framework_support(self, framework: Union[str, Framework]):
        """
        Test if the framework is supported by this task.
        """
        if isinstance(framework, str):
            framework = Framework(framework)
        if framework not in self.supported_frameworks:
            raise NotImplementedError(f"Framework {framework} is not supported for this {self.__class__}.")

    def _execute_formatting_func(self, dataset: "FeedbackDataset") -> Any:
        """
        Execute the formatting function on the dataset and return the output.
        """

        def test_none_sample(sample: Any) -> list:
            """
            Check for None values in the sample. If there are None values, return an empty list.
            """

            if sample is None:
                return []
            if isinstance(sample, (list, typing.Generator)):
                values = [test_none_sample(entry) for entry in sample]
                values = [value for value in values if value]
                values = [value[0] for value in values]
                return values
            else:
                return [sample]

        def test_output_formatting_func(sample: Any) -> list:
            """
            Test if the formatting function returns the expected format.
            """
            try:
                if isinstance(sample, list):
                    return [self._formatting_func_return_types(format=entry).format for entry in sample]
                else:
                    return [self._formatting_func_return_types(format=sample).format]
            except Exception:
                raise ValueError(
                    f"formatting_func must return {self._formatting_func_return_types.__annotations__['format']}, not {type(sample)}"
                )

        formatted_output = []
        for sample in dataset.format_as("datasets"):
            formatted_sample = self.formatting_func(sample)
            formatted_sample = test_none_sample(formatted_sample)
            if not formatted_sample:
                continue
            formatted_output += test_output_formatting_func(formatted_sample)
        return formatted_output

    def _format_data(self, dataset: "FeedbackDataset") -> List[Dict[str, Any]]:
        formatted_data = []
        explode_columns = set()
        for record in dataset.records:
            data = {}
            for pydantic_field in self.defaults:
                # with default and formatting_func either one can be None
                if pydantic_field[-1] is not None:
                    pydantic_field_name, pydantic_field_value = pydantic_field
                    if isinstance(pydantic_field_value, (list, tuple, set)):
                        # NOTE: In the case of TrainingTaskForSentenceSimilarity with defaults,
                        # we need to grab multiple values, not just the text.
                        for pydantic_field_value_i in pydantic_field_value:
                            if isinstance(pydantic_field_value_i, (TextField,)):
                                data[pydantic_field_value_i.name] = record.fields[pydantic_field_value_i.name]
                            else:
                                if pydantic_field_value_i.question.name not in record._unified_responses:
                                    continue
                                else:
                                    data[pydantic_field_name] = [
                                        resp.value
                                        for resp in record._unified_responses[pydantic_field_value_i.question.name]
                                    ]
                                explode_columns.add(pydantic_field_name)

                    elif isinstance(pydantic_field_value, (TextField,)):
                        data[pydantic_field_name] = record.fields[pydantic_field_value.name]
                    elif isinstance(pydantic_field_value, (TextQuestion,)):
                        if pydantic_field_value.name not in record._unified_responses:
                            continue
                        else:
                            data[pydantic_field_name] = record._unified_responses[pydantic_field_value.name].value
                        explode_columns.add(pydantic_field_name)
                    else:
                        if pydantic_field_value.question.name not in record._unified_responses:
                            continue
                        else:
                            data[pydantic_field_name] = [
                                resp.value for resp in record._unified_responses[pydantic_field_value.question.name]
                            ]
                        explode_columns.add(pydantic_field_name)
            formatted_data.append(data)
        df = pd.DataFrame(formatted_data)
        if explode_columns:
            df = df.explode(list(explode_columns))
        # In cases of MultiLabel datasets the label column contains a list,
        # which is unhashable, so for those cases we transform the rows in the
        # dataframe to tuples to drop duplicates and reconstruct the original
        # dataframe format.
        if df.applymap(lambda x: isinstance(x, list)).any().any():
            df = pd.DataFrame(df.apply(tuple, 1).drop_duplicates().to_list(), columns=df.columns)
        else:
            df = df.drop_duplicates()
        df = df.dropna(how="any")
        records = df.to_dict(orient="records")

        # Validate record format
        if isinstance(self.defaults, SentenceSimilarityDefaults):
            validated_records = []
            for rec in records:
                try:
                    self._formatting_func_return_types(format=rec)
                    validated_records.append(rec)
                except Exception:
                    continue
        elif isinstance(self.defaults, (QuestionAnsweringDefaults, TextClassificationDefaults)):
            required_keys = list(self.defaults.__fields__.keys())
            validated_records = [item for item in records if all(key in item for key in required_keys)]
        else:
            raise NotImplementedError(
                f"Defaults {self.defaults} is not supported. Choose from: {[e.value for e in Framework]}"
            )
        if not validated_records:
            raise ValueError(
                f"Your dataset does not contain any records with required responses for {required_keys}. "
                "Try different `filter_by`and `max_records` params or a annotate some more records."
            )
        return records

    def _train_test_split(self, data: List[dict], train_size: float, seed: int) -> Tuple[List[dict], List[dict]]:
        """Overwritten by subclasses"""

    def prepare_for_training(
        self, framework: Framework, dataset: "FeedbackDataset", train_size: float, seed: int, lang: str
    ) -> Any:
        data = self._format_data(dataset)
        if framework in [
            Framework.TRANSFORMERS,
            Framework.SETFIT,
            Framework.SPAN_MARKER,
            Framework.PEFT,
        ]:
            return self._prepare_for_training_with_transformers(
                data=data, train_size=train_size, seed=seed, framework=framework
            )
        elif framework in [Framework.SPACY, Framework.SPACY_TRANSFORMERS]:
            require_dependencies("spacy")
            import spacy

            if lang is None:
                _LOGGER.warning("spaCy `lang` is not provided. Using `en`(English) as default language.")
                lang = spacy.blank("en")
            elif lang.isinstance(str):
                if len(lang) == 2:
                    lang = spacy.blank(lang)
                else:
                    lang = spacy.load(lang)
            return self._prepare_for_training_with_spacy(data=data, train_size=train_size, seed=seed, lang=lang)
        elif framework is Framework.SPARK_NLP:
            return self._prepare_for_training_with_spark_nlp(data=data, train_size=train_size, seed=seed)
        elif framework is Framework.OPENAI:
            return self._prepare_for_training_with_openai(data=data, train_size=train_size, seed=seed)
        elif framework is Framework.TRL:
            return self._prepare_for_training_with_trl(data=data, train_size=train_size, seed=seed)

        elif framework is Framework.SENTENCE_TRANSFORMERS:
            return self._prepare_for_training_with_sentence_transformers(data=data, train_size=train_size, seed=seed)
        else:
            raise NotImplementedError(
                f"Framework {framework} is not supported. Choose from: {[e.value for e in Framework]}"
            )

    def _prepare_for_training_with_transformers(
        self, data: List[dict], train_size, seed: int, framework: Union[str, Framework]
    ) -> Union["datasets.Dataset", "datasets.DatasetDict"]:
        raise ValueError(f"{self.__class__.__name__} does not support the {framework} framework.")

    def _prepare_for_training_with_spacy(
        self, data: List[dict], train_size, seed: int, lang: str
    ) -> Union["spacy.token.DocBin", Tuple["spacy.token.DocBin", "spacy.token.DocBin"]]:
        raise ValueError(f"{self.__class__.__name__} does not support the spaCy framework.")

    def _prepare_for_training_with_spark_nlp(
        self, data: List[dict], train_size, seed: int
    ) -> Union["pd.DataFrame", Tuple["pd.DataFrame", "pd.DataFrame"]]:
        raise ValueError(f"{self.__class__.__name__} does not support the Spark NLP framework.")

    def _prepare_for_training_with_openai(
        self, data: List[dict], train_size, seed: int
    ) -> Union[List[dict], Tuple[List[dict], List[dict]]]:
        raise ValueError(f"{self.__class__.__name__} does not support the OpenAI framework.")

    def _prepare_for_training_with_trl(
        self, data: List[dict], train_size, seed: int
    ) -> Union[List[dict], Tuple[List[dict], List[dict]]]:
        raise ValueError(f"{self.__class__.__name__} does not support the TRL framework.")


class TrainingTask:
    @classmethod
    def for_text_classification(
        cls,
        formatting_func: Callable[[Dict[str, Any]], Union[None, str, List[str], Iterator[str]]] = None,
        text: Optional[TextField] = None,
        label: Optional[
            Union[
                RatingQuestion,
                LabelQuestion,
                RankingQuestion,
                MultiLabelQuestion,
                RatingQuestionUnification,
                LabelQuestionUnification,
                MultiLabelQuestionUnification,
                RankingQuestionUnification,
            ]
        ] = None,
        label_strategy: str = None,
    ) -> "TrainingTaskForTextClassification":
        """
        Define a task configuration for text classification. It takes default values for `text` and `label` using datasets Fields and Questions or a custom `formatting_func` as Callable. See Examples underneath for more details.

        Args:
            formatting_func: A formatting function. Defaults to None.
            text: The TextField to use for training. Defaults to None.
            label: The *Question to use for training. Defaults to None.
            label_strategy: A strategy to unify responses. Defaults to None. This means it will initialize the default strategy for the label type. Defaults to None.

        Raises:
            ValueError: if label is not a valid type with the question type.
            ValueError: if label_strategy is defined and label is already a Unification class.

        Returns:
            TrainingTaskForTextClassification: A task mapping instance to be used in `FeedbackDataset.prepare_for_training()`

        Examples:
            >>> # with defaults
            >>> from argilla import LabelQuestion, TrainingTask
            >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
            >>> task = TrainingTask.for_text_classification(
            ...     text=dataset.field_by_name("text"),
            ...     label=dataset.question_by_name("label")
            ... )
            >>> dataset.prepare_for_training(framework="...", task=task)
            >>> # with formatting_func
            >>> from argilla import LabelQuestion, TrainingTask
            >>> from collections import Counter
            >>> import random
            >>> def formatting_func(sample: Dict[str, Any]) -> Union[Tuple[str, str], Tuple[str, List[str]]]:
            ...     text = sample["text"]
            ...     values = [annotation["value"] for annotation in sample["label"]]
            ...     counter = Counter(values)
            ...     if counter:
            ...         most_common = counter.most_common()
            ...         max_frequency = most_common[0][1]
            ...         most_common_elements = [element for element, frequency in most_common if frequency == max_frequency]
            ...         label = random.choice(most_common_elements)
            ...         return (text, label)
            >>> task = TrainingTask.for_text_classification(formatting_func=formatting_func)
            >>> dataset.prepare_for_training(framework="...", task=task)
        """
        if (text and label) and formatting_func is not None:
            raise ValueError("You must provide either `text` and `label`, or a `formatting_func`, not both.")

        if formatting_func is not None:
            if text or label:
                raise ValueError("`formatting_func` is already defined, so you cannot define `text` and `label`.")
            return TrainingTaskForTextClassification(formatting_func=formatting_func)
        else:
            if isinstance(
                label,
                (
                    LabelQuestionUnification,
                    MultiLabelQuestionUnification,
                    RatingQuestionUnification,
                    RankingQuestionUnification,
                ),
            ):
                if label_strategy is not None:
                    raise ValueError("label_strategy is already defined via Unification class.")
            else:
                unification_kwargs = {"question": label}
                if label_strategy is not None:
                    unification_kwargs["strategy"] = label_strategy
                else:
                    _LOGGER.info(f"No label strategy defined. Using default strategy for {type(label)}.")
                if isinstance(label, RatingQuestion):
                    label = RatingQuestionUnification(**unification_kwargs)
                elif isinstance(label, MultiLabelQuestion):
                    label = MultiLabelQuestionUnification(**unification_kwargs)
                elif isinstance(label, LabelQuestion):
                    label = LabelQuestionUnification(**unification_kwargs)
                elif isinstance(label, RankingQuestion):
                    label = RankingQuestionUnification(**unification_kwargs)
                else:
                    raise ValueError(f"Label type {type(label)} is not supported.")
            defaults = TextClassificationDefaults(text=text, label=label)
            return TrainingTaskForTextClassification(defaults=defaults)

    @classmethod
    def for_supervised_fine_tuning(
        cls,
        formatting_func: Callable[[Dict[str, Any]], Union[None, str, List[str], Iterator[str]]],
    ) -> "TrainingTaskForSFT":
        """
        Return a task that can be used in `FeedbackDataset.prepare_for_training(framework="...", task)`
        to extract data from the Feedback Dataset in an immediately useful format.

        Args:
            formatting_func: A formatting function converting a dictionary of records into zero,
                one or more text strings.

        Returns:
            TrainingTaskForSFT: A task mapping instance to be used in `FeedbackDataset.prepare_for_training()`

        Examples:
            >>> from argilla import TrainingTask
            >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
            >>> def formatting_func(sample: Dict[str, Any]):
            ...     annotations = sample["good]
            ...     if annotations and annotations[0]["value"] == "Bad":
            ...         return
            ...     return template.format(prompt=sample["prompt"][0]["value"], response=sample["response"][0]["value"])
            >>> task = TrainingTask.for_supervised_fine_tuning(formatting_func=formatting_func)
            >>> dataset.prepare_for_training(framework="...", task=task)

        """
        return TrainingTaskForSFT(formatting_func=formatting_func)

    @classmethod
    def for_reward_modeling(
        cls,
        formatting_func: Callable[
            [Dict[str, Any]], Union[None, Tuple[str, str], List[Tuple[str, str]], Iterator[Tuple[str, str]]]
        ],
    ) -> "TrainingTaskForRM":
        """
        Return a task that can be used in `FeedbackDataset.prepare_for_training(framework="...", task)`
        to extract data from the Feedback Dataset in an immediately useful format.

        Args:
            formatting_func: A formatting function converting a dictionary of records into zero,
                one or more chosen-rejected text tuples.

        Returns:
            TrainingTaskForRM: A task mapping instance to be used in `FeedbackDataset.prepare_for_training()`

        Examples:
            >>> from argilla import TrainingTask
            >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
            >>> def formatting_func(sample: Dict[str, Any]):
            ...     values = [annotation["value"] for annotation in sample["ranking"]]
            ...     if values.count("1") >= values.count("2"):
            ...         chosen = sample["response-1"]
            ...         rejected = sample["response-2"]
            ...     else:
            ...         chosen = sample["response-2"]
            ...         rejected = sample["response-1"]
            ...     yield chosen, rejected
            >>> task = TrainingTask.for_reward_modeling(formatting_func=formatting_func)
            >>> dataset.prepare_for_training(framework="...", task=task)

        """
        return TrainingTaskForRM(formatting_func=formatting_func)

    @classmethod
    def for_proximal_policy_optimization(
        cls, formatting_func: Callable[[Dict[str, Any]], Union[None, str, Iterator[str]]]
    ) -> "TrainingTaskForPPO":
        """
        Return a task that can be used in `FeedbackDataset.prepare_for_training(text: TextField)`
        to extract data from the Feedback Dataset in an immediately useful format.

        Args:
            formatting_func: A formatting function converting a dictionary of records into zero,
                one or more prompts.

        Returns:
            TrainingTaskForPPO: A task mapping instance to be used in `FeedbackDataset.prepare_for_training()`
        """
        return TrainingTaskForPPO(formatting_func=formatting_func)

    @classmethod
    def for_direct_preference_optimization(
        cls,
        formatting_func: Callable[[Dict[str, Any]], Union[None, Tuple[str, str, str], Iterator[Tuple[str, str, str]]]],
    ) -> "TrainingTaskForDPO":
        """
        Provide `TrainingTask.for_direct_preference_optimization(formatting_func: Callable)`
        Return a task that can be used in `FeedbackDataset.prepare_for_training(framework="...", task)`
        to extract data from the Feedback Dataset in an immediately useful format.

        Args:
            formatting_func: A formatting function converting a dictionary of records into zero,
                one or more prompt-chosen-rejected text tuples.

        Returns:
            TrainingTaskForDPO: A task mapping instance to be used in `FeedbackDataset.prepare_for_training()`

        Examples:
            >>> from argilla import TrainingTask
            >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
            >>> def formatting_func(sample: Dict[str, Any]):
            ...     values = [annotation["value"] for annotation in sample["ranking"]]
            ...     if values.count("1") >= values.count("2"):
            ...         chosen = sample["response-1"]
            ...         rejected = sample["response-2"]
            ...     else:
            ...         chosen = sample["response-2"]
            ...         rejected = sample["response-1"]
            ...     yield sample["prompt"], chosen, rejected
            >>> task = TrainingTask.for_direct_preference_optimization(formatting_func=formatting_func)
            >>> dataset.prepare_for_training(framework="...", task=task)

        """
        return TrainingTaskForDPO(formatting_func=formatting_func)

    @classmethod
    def for_chat_completion(
        cls,
        formatting_func: Callable[
            [Dict[str, Any]], Union[None, Tuple[str, str, str, str], Iterator[Tuple[str, str, str, str]]]
        ],
    ) -> "TrainingTaskForChatCompletion":
        """Training data for chat completion
        Args:
            formatting_func: A formatting function converting a dictionary of records into zero,
                one or more chat-turn-role-content text tuples.

        Examples:
            >>> from argilla import TrainingTaskForChatCompletion
            >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
            >>> def formatting_func(sample: Dict[str, Any]):
            ...     from uuid import uuid4
            ...     chat_id = str(uuid4())
            ...     if sample["response"]:
            ...         chat = str(uuid4())
            ...         user_message = sample["prompt"][0]["value"]
            ...         system_message = sample["response"][0]["value"]
            ...         yield [(chat, "0", "user", user_message), (chat, "1", "assistant", system_message)]
            >>> task = TrainingTaskForChatCompletion(formatting_func=formatting_func)
            >>> dataset.prepare_for_training(framework="...", task=task)
        """
        return TrainingTaskForChatCompletion(formatting_func=formatting_func)

    @classmethod
    def for_question_answering(
        cls,
        formatting_func: Optional[
            Callable[[Dict[str, Any]], Union[None, Tuple[str, str, str], Iterator[Tuple[str, str, str]]]]
        ] = None,
        question: Optional[TextField] = None,
        context: Optional[TextField] = None,
        answer: Optional[TextQuestion] = None,
    ) -> "TrainingTaskForQuestionAnswering":
        """Training data for question answering

        Args:
            formatting_func: A formatting function converting a dictionary of records into zero,
                one or more question-context-answer text tuples.
            question: The TextField to use for training.
            context: The TextField to use for training.
            answer: The TextQuestion to use for training.

        Examples:
            >>> # with defaults
            >>> from argilla import TrainingTaskForQuestionAnswering
            >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
            >>> task = TrainingTaskForQuestionAnswering(
            ...     question=dataset.field_by_name("question"),
            ...     context=dataset.field_by_name("context"),
            ...     answer=dataset.question_by_name("answer"),
            ... )
            >>> dataset.prepare_for_training(framework="...", task=task)
            >>> # with formatting_func
            >>> from argilla import TrainingTaskForQuestionAnswering
            >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
            >>> def formatting_func(sample: Dict[str, Any]):
            ...     question = sample["question"]
            ...     context = sample["context"]
            ...     for answer in sample["answer"]:
            ...         if not all([question, context, answer["value"]]):
            ...             continue
            ...         yield question, context, answer["value"]
            >>> task = TrainingTaskForQuestionAnswering(formatting_func=formatting_func)
            >>> dataset.prepare_for_training(framework="...", task=task)
        """
        if (question and context and answer) and formatting_func is not None:
            raise ValueError(
                "You must provide either `question`, `context` and `answer`, or a `formatting_func`, not both."
            )

        if formatting_func is not None:
            if question or context or answer:
                raise ValueError(
                    "`formatting_func` is already defined, so you cannot define `question`, `context` and `answer`."
                )
            return TrainingTaskForQuestionAnswering(formatting_func=formatting_func)
        else:
            defaults = QuestionAnsweringDefaults(question=question, context=context, answer=answer)
            return TrainingTaskForQuestionAnswering(defaults=defaults)

    @classmethod
    def for_sentence_similarity(
        cls,
        texts: Optional[List[TextField]] = None,
        label: Optional[Union[LabelQuestion, RankingQuestion]] = None,
        formatting_func: Callable[
            [Dict[str, Any]],
            Union[
                None,
                Dict[str, Union[float, int]],
                Dict[str, str],
                List[Dict[str, Union[float, int]]],
                List[Dict[str, str]],
            ],
        ] = None,
        label_strategy: Optional[Union[LabelQuestionUnification, RatingQuestionUnification]] = None,
    ) -> "TrainingTaskForSentenceSimilarity":
        """

        Return a task that can be used in `FeedbackDataset.prepare_for_training(framework="...", task)`
        to extract data from the Feedback Dataset in a format suitable for sentence similarity.

        Args:
            texts: A list of TextFields to use for training, typically two text pieces, can be a triplet also.
                Defaults to None.
            label: The `LabelQuestion` or `RankingQuestion` to use for training. These models can be trained without
                explicit use of labels, just with pairs or triplets of texts. Defaults to None.
            formatting_func: A formatting function converting a dictionary of records into a dict
                of `sentence-1`-`sentence-2` pairs or triplets `sentence-1`-`sentence-2`-`sentence-3`,
                optionally including a `label` field.
            label_strategy: A strategy to unify responses. Defaults to None. This means it will initialize the default strategy for the label type.

        Raises:
            ValueError: If label is not a valid type.
            ValueError: if label_strategy is defined and label is already a Unification class.

        Returns:
            TrainingTaskForSentenceSimilarity: A task mapping instance to be used in `FeedbackDataset.prepare_for_training()`

        Examples:
            >>> from argilla import LabelQuestion, TrainingTask
            >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
            >>> task = TrainingTask.for_text_classification(
            ...     texts=[dataset.field_by_name("premise"), dataset.field_by_name("hypothesis")],
            ...     label=dataset.question_by_name("label")
            ... )
            >>> dataset.prepare_for_training(framework="...", task=task)

            >>> from argilla import LabelQuestion, TrainingTask
            >>> from collections import Counter
            >>> import random
            >>> def formatting_func(sample: Dict[str, Any]) -> Union[Tuple[str, str], Tuple[str, List[str]]]:
            ...     record = {"sentence-1": sample["premise"], "sentence-2": sample["hypothesis"]}
            ...     values = [annotation["value"] for annotation in sample["label"]]
            ...     counter = Counter(values)
            ...     if counter:
            ...         most_common = counter.most_common()
            ...         max_frequency = most_common[0][1]
            ...         most_common_elements = [element for element, frequency in most_common if frequency == max_frequency]
            ...         record["label"] = label
            ...         return record
            ...     else:
            ...         return None
            >>> task = TrainingTask.for_sentence_similarity(formatting_func=formatting_func)
            >>> dataset.prepare_for_training(framework="...", task=task)
        """

        if (texts or label) and formatting_func is not None:
            raise ValueError(
                "You must provide either `texts` and (optionally) `label`, or a `formatting_func`, not both."
            )

        if formatting_func is not None:
            return TrainingTaskForSentenceSimilarity(formatting_func=formatting_func)
        else:
            if not label:
                defaults = SentenceSimilarityDefaults(texts=texts)
                return TrainingTaskForSentenceSimilarity(defaults=defaults)

            if isinstance(label, LabelQuestionUnification):
                if label_strategy is not None:
                    raise ValueError("label_strategy is already defined via Unification class.")
            else:
                unification_kwargs = {"question": label}
                if label_strategy is not None:
                    unification_kwargs["strategy"] = label_strategy
                else:
                    _LOGGER.info(f"No label strategy defined. Using default strategy for {type(label)}.")
                if isinstance(label, LabelQuestion):
                    label = LabelQuestionUnification(**unification_kwargs)
                elif isinstance(label, RatingQuestion):
                    label = RatingQuestionUnification(**unification_kwargs)
                else:
                    raise ValueError(f"Label type {type(label)} is not supported.")
            defaults = SentenceSimilarityDefaults(texts=texts, label=label)
            return TrainingTaskForSentenceSimilarity(defaults=defaults)


class TrainingTaskForTextClassification(BaseModel, TrainingData):
    """Training data for text classification

    Args:
        formatting_func: A formatting function returning the text to classify. Either a formatting function or
            the text and label parameters are provided. Defaults to None.
        text: The text field to take as the text to classify.
        label: The question denoting the label of the text to classify.

        Examples:
            >>> # with defaults
            >>> from argilla import LabelQuestion, TrainingTask
            >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
            >>> task = TrainingTask.for_text_classification(
            ...     text=dataset.field_by_name("text"),
            ...     label=dataset.question_by_name("label")
            ... )
            >>> dataset.prepare_for_training(framework="...", task=task)
            >>> # with formatting_func
            >>> from argilla import LabelQuestion, TrainingTask
            >>> from collections import Counter
            >>> import random
            >>> def formatting_func(sample: Dict[str, Any]) -> Union[Tuple[str, str], Tuple[str, List[str]]]:
            ...     text = sample["text"]
            ...     values = [annotation["value"] for annotation in sample["label"]]
            ...     counter = Counter(values)
            ...     if counter:
            ...         most_common = counter.most_common()
            ...         max_frequency = most_common[0][1]
            ...         most_common_elements = [element for element, frequency in most_common if frequency == max_frequency]
            ...         label = random.choice(most_common_elements)
            ...         yield text, label
            >>> task = TrainingTask.for_text_classification(formatting_func=formatting_func)
            >>> dataset.prepare_for_training(framework="...", task=task)
    """

    formatting_func: Optional[Callable[[Dict[str, Any]], Union[None, str, List[str], Iterator[str]]]] = None
    defaults: Optional[TextClassificationDefaults] = TextClassificationDefaults()
    _formatting_func_return_types = TextClassificationReturnTypes
    _supported_frameworks_names = [
        "transformers",
        "spacy",
        "openai",
        "setfit",
        "peft",
        "spark-nlp",
        "spacy-transformers",
    ]

    @property
    def __multi_label__(self) -> bool:
        return isinstance(self.label.question, MultiLabelQuestion)

    @property
    def __all_labels__(self) -> Union[Any, List[str]]:
        return self.label.question.__all_labels__

    @property
    def __label2id__(self) -> Union[Any, Dict[str, int]]:
        return self.label.question.__label2id__

    @property
    def __id2label__(self) -> Dict[int, str]:
        return self.label.question.__id2label__

    @property
    def label(
        self,
    ) -> Optional[
        Union[
            RatingQuestion,
            LabelQuestion,
            MultiLabelQuestion,
            RankingQuestion,
            LabelQuestionUnification,
            MultiLabelQuestionUnification,
            RankingQuestionUnification,
            RatingQuestionUnification,
        ]
    ]:
        return self.defaults.label

    @property
    def text(self) -> Optional[TextField]:
        return self.defaults.text

    def _format_data(self, dataset: "FeedbackDataset") -> List[Dict[str, Any]]:
        if self.formatting_func is not None:
            output = self._execute_formatting_func(dataset)

            data = []
            _all_labels = set()
            for text_label in output:
                text, label = text_label
                data.append({"text": text, "label": label})
                if isinstance(label, list):
                    _multi_label = True
                    _all_labels |= set(label)
                else:
                    _all_labels.add(label)
                    _multi_label = False

            # infer label type from output custom formatting function
            if _multi_label:
                self.defaults.label = MultiLabelQuestionUnification(
                    question=MultiLabelQuestion(name="custom_func", labels=list(_all_labels))
                )
            else:
                self.defaults.label = LabelQuestionUnification(
                    question=LabelQuestion(name="custom_func", labels=list(_all_labels))
                )
            return data
        else:
            return super()._format_data(dataset)

    def compute_unified_responses(self, responses: List[FeedbackRecord]):
        self.defaults.label.strategy.compute_unified_responses(responses=responses, field=self.defaults.label.question)

    @requires_dependencies("scikit-learn")
    def _train_test_split(self, data: List[dict], train_size: float, seed: int) -> Tuple[List[dict], List[dict]]:
        from sklearn.model_selection import train_test_split

        # TODO: Stratify by label
        # TODO: provide label overview
        return train_test_split(
            data,
            train_size=train_size,
            shuffle=True,
            random_state=seed,
        )

    @requires_dependencies("datasets>1.17.0")
    def _prepare_for_training_with_transformers(
        self, data: List[dict], train_size: float, seed: int, framework: Union[str, Framework]
    ) -> Union["datasets.Dataset", "datasets.DatasetDict"]:
        self.test_framework_support(framework)
        import datasets

        multi_label = self.__multi_label__

        datasets_dict = {"id": [], "text": [], "label": []}

        for index, entry in enumerate(data):
            datasets_dict["id"].append(index)
            datasets_dict["text"].append(entry["text"])
            datasets_dict["label"].append(entry["label"])

        all_labels = self.label.question.__all_labels__
        class_label = datasets.ClassLabel(names=all_labels)
        feature_dict = {
            "id": datasets.Value(dtype="int32"),
            "text": datasets.Value("string"),
            "label": [class_label] if multi_label else class_label,
        }

        ds = datasets.Dataset.from_dict(datasets_dict, features=datasets.Features(feature_dict))

        if multi_label:
            require_dependencies("scikit-learn")
            from sklearn.preprocessing import MultiLabelBinarizer

            labels = [rec["label"] for rec in ds]
            mlb = MultiLabelBinarizer()
            binarized_labels = mlb.fit_transform(labels)
            feature_dict["binarized_label"] = feature_dict["label"]
            ds = datasets.Dataset.from_dict(
                {
                    "id": ds["id"],
                    "text": ds["text"],
                    "label": labels,
                    "binarized_label": binarized_labels,
                },
                features=datasets.Features(feature_dict),
            )
        if train_size != 1:
            ds = ds.train_test_split(train_size=train_size, test_size=1 - train_size, seed=seed)

        return ds

    @requires_dependencies("spacy")
    def _prepare_for_training_with_spacy(
        self, data: List[dict], train_size: float, seed: int, lang: str
    ) -> Union["spacy.token.DocBin", Tuple["spacy.token.DocBin", "spacy.token.DocBin"]]:
        from spacy.tokens import DocBin

        all_labels = self.__all_labels__

        def _prepare(data):
            db = DocBin(store_user_data=True)
            # Creating the DocBin object as in https://spacy.io/usage/training#training-data
            for entry in data:
                doc = lang.make_doc(entry["text"])

                cats = dict.fromkeys(all_labels, 0)
                if isinstance(entry["label"], list):
                    for label in entry["label"]:
                        cats[label] = 1
                else:
                    cats[entry["label"]] = 1

                doc.cats = cats
                db.add(doc)
            return db

        if train_size != 1:
            train_data, test_data = self._train_test_split(data, train_size, seed)
            return _prepare(train_data), _prepare(test_data)
        else:
            return _prepare(data)

    def _prepare_for_training_with_spark_nlp(
        self, data: List[dict], train_size: float, seed: int
    ) -> Union["pd.DataFrame", Tuple["pd.DataFrame", "pd.DataFrame"]]:
        def _prepare(data):
            df = pd.DataFrame(data)
            if multi_label:
                df.rename(columns={"label": "labels"}, inplace=True)
            return df

        multi_label = isinstance(self.label.question, MultiLabelQuestion)
        if train_size != 1:
            train_data, test_data = self._train_test_split(data, train_size, seed)
            return _prepare(train_data), _prepare(test_data)
        else:
            return _prepare(data)

    def _prepare_for_training_with_openai(
        self, data: List[dict], train_size: float, seed: int
    ) -> Union[List[dict], Tuple[List[dict], List[dict]]]:
        separator = OPENAI_SEPARATOR
        whitespace = OPENAI_WHITESPACE
        label2id = self.__label2id__

        if len(data) * train_size <= len(label2id) * 100:
            _LOGGER.warning("OpenAI recommends at least 100 examples per class for training a classification model.")

        def _prepare(data):
            jsonl = []
            for entry in data:
                prompt = entry["text"]
                prompt += separator  # needed for better performance

                try:
                    if multi_label:
                        completion = " ".join([str(label2id[str(annotation)]) for annotation in entry["label"]])
                    else:
                        completion = str(label2id[str(entry["label"])])
                except KeyError:
                    if multi_label:
                        completion = " ".join([str(label2id[int(annotation)]) for annotation in entry["label"]])
                    else:
                        completion = str(label2id[int(entry["label"])])

                jsonl.append(
                    {
                        # "id": rec.id,
                        "prompt": prompt,
                        "completion": whitespace + completion,
                    }
                )

            return jsonl

        multi_label = isinstance(self.label.question, MultiLabelQuestion)
        if train_size != 1:
            train_data, test_data = self._train_test_split(data, train_size, seed)
            return _prepare(train_data), _prepare(test_data)
        else:
            return _prepare(data)


class TrainingTaskForSFT(BaseModel, TrainingData):
    """Training data for supervised finetuning

    Args:
        formatting_func: A formatting function converting a dictionary of records into zero,
            one or more text strings.

    Examples:
        >>> from argilla import TrainingTaskForSFT
        >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
        >>> def formatting_func(sample: Dict[str, Any]):
        ...     annotations = sample["good]
        ...     if annotations and annotations[0]["value"] == "Bad":
        ...         return
        ...     yield template.format(prompt=sample["prompt"][0]["value"], response=sample["response"][0]["value"])
        >>> task = TrainingTaskForSFT(formatting_func=formatting_func)
        >>> dataset.prepare_for_training(framework="...", task=task)

    """

    formatting_func: Callable[[Dict[str, Any]], Union[None, str, List[str], Iterator[str]]]
    _formatting_func_return_types = SFTReturnTypes
    _supported_frameworks_names = ["trl"]

    def _format_data(self, dataset: "FeedbackDataset") -> List[Dict[str, str]]:
        formatted_output = self._execute_formatting_func(dataset)
        return [{"text": text} for text in formatted_output]

    @requires_dependencies("datasets>1.17.0")
    def _prepare_for_training_with_trl(
        self, data: List[dict], train_size: float, seed: int
    ) -> Union["datasets.Dataset", "datasets.DatasetDict"]:
        import datasets

        datasets_dict = {"id": [], "text": []}
        for index, sample in enumerate(data):
            datasets_dict["id"].append(index)
            datasets_dict["text"].append(sample["text"])

        feature_dict = {
            "id": datasets.Value(dtype="int32"),
            "text": datasets.Value("string"),
        }

        ds = datasets.Dataset.from_dict(datasets_dict, features=datasets.Features(feature_dict))
        if train_size != 1:
            ds = ds.train_test_split(train_size=train_size, test_size=1 - train_size, seed=seed)

        return ds


class TrainingTaskForRM(BaseModel, TrainingData):
    """Training data for reward modeling

    Args:
        formatting_func: A formatting function converting a dictionary of records into zero,
            one or more chosen-rejected text tuples.

    Examples:
        >>> from argilla import TrainingTaskForRM
        >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
        >>> def formatting_func(sample: Dict[str, Any]):
        ...     values = [annotation["value"] for annotation in sample["ranking"]]
        ...     if values.count("1") >= values.count("2"):
        ...         chosen = sample["response-1"]
        ...         rejected = sample["response-2"]
        ...     else:
        ...         chosen = sample["response-2"]
        ...         rejected = sample["response-1"]
        ...     yield chosen, rejected
        >>> task = TrainingTaskForRM(formatting_func=formatting_func)
        >>> dataset.prepare_for_training(framework="...", task=task)
    """

    formatting_func: Callable[
        [Dict[str, Any]], Union[None, Tuple[str, str], List[Tuple[str, str]], Iterator[Tuple[str, str]]]
    ]
    _formatting_func_return_types = RMReturnTypes
    _supported_frameworks_names = ["trl"]

    def _format_data(self, dataset: "FeedbackDataset") -> List[Dict[str, str]]:
        output = self._execute_formatting_func(dataset)
        return [{"chosen": chosen, "rejected": rejected} for chosen, rejected in output]

    @requires_dependencies("datasets>1.17.0")
    def _prepare_for_training_with_trl(
        self, data: List[dict], train_size: float, seed: int
    ) -> Union["datasets.Dataset", "datasets.DatasetDict"]:
        import datasets

        datasets_dict = {"chosen": [], "rejected": []}
        for sample in data:
            datasets_dict["chosen"].append(sample["chosen"])
            datasets_dict["rejected"].append(sample["rejected"])

        feature_dict = {
            "rejected": datasets.Value("string"),
            "chosen": datasets.Value("string"),
        }

        ds = datasets.Dataset.from_dict(datasets_dict, features=datasets.Features(feature_dict))
        if train_size != 1:
            ds = ds.train_test_split(train_size=train_size, test_size=1 - train_size, seed=seed)

        return ds


class TrainingTaskForPPO(BaseModel, TrainingData):
    """Training data for proximal policy optimization

    Args:
        text: The TextField to use for training.

    Examples:
        >>> from argilla import TrainingTaskForPPO
        >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
        >>> task = TrainingTaskForPPO(text=dataset.fields[0],)
        >>> dataset.prepare_for_training(framework="...", task=task)
    """

    formatting_func: Callable[[Dict[str, Any]], Union[None, str, Iterator[str]]]
    _formatting_func_return_types = PPOReturnTypes
    _supported_frameworks_names = ["trl"]

    def _format_data(self, dataset: "FeedbackDataset") -> List[Dict[str, str]]:
        output = self._execute_formatting_func(dataset)
        return [{"query": text} for text in output]

    @requires_dependencies("datasets>1.17.0")
    def _prepare_for_training_with_trl(
        self, data: List[dict], train_size: float, seed: int
    ) -> Union["datasets.Dataset", "datasets.DatasetDict"]:
        import datasets

        datasets_dict = {"id": [], "query": []}
        for index, entry in enumerate(data):
            datasets_dict["id"].append(index)
            datasets_dict["query"].append(entry["query"])

        feature_dict = {
            "id": datasets.Value(dtype="int32"),
            "query": datasets.Value("string"),
        }

        ds = datasets.Dataset.from_dict(datasets_dict, features=datasets.Features(feature_dict))

        if train_size != 1:
            ds = ds.train_test_split(train_size=train_size, test_size=1 - train_size, seed=seed)

        return ds


class TrainingTaskForDPO(BaseModel, TrainingData):
    """Training data for direct preference optimization

    Args:
        formatting_func: A formatting function converting a dictionary of records into zero,
            one or more prompt-chosen-rejected text tuples.

    Examples:
        >>> from argilla import TrainingTaskForDPO
        >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
        >>> def formatting_func(sample: Dict[str, Any]):
        ...     values = [annotation["value"] for annotation in sample["ranking"]]
        ...     if values.count("1") >= values.count("2"):
        ...         chosen = sample["response-1"]
        ...         rejected = sample["response-2"]
        ...     else:
        ...         chosen = sample["response-2"]
        ...         rejected = sample["response-1"]
        ...     yield sample["prompt"], chosen, rejected
        >>> task = TrainingTaskForDPO(formatting_func=formatting_func)
        >>> dataset.prepare_for_training(framework="...", task=task)
    """

    formatting_func: Callable[[Dict[str, Any]], Union[None, Tuple[str, str, str], Iterator[Tuple[str, str, str]]]]
    _formatting_func_return_types = DPOReturnTypes
    _supported_frameworks_names = ["trl"]

    def _format_data(self, dataset: "FeedbackDataset") -> List[Dict[str, str]]:
        output = self._execute_formatting_func(dataset)
        return [{"prompt": prompt, "chosen": chosen, "rejected": rejected} for prompt, chosen, rejected in output]

    @requires_dependencies("datasets>1.17.0")
    def _prepare_for_training_with_trl(
        self, data: List[dict], train_size: float, seed: int
    ) -> Union["datasets.Dataset", "datasets.DatasetDict"]:
        import datasets

        datasets_dict = {"prompt": [], "chosen": [], "rejected": []}
        for sample in data:
            datasets_dict["prompt"].append(sample["prompt"])
            datasets_dict["chosen"].append(sample["chosen"])
            datasets_dict["rejected"].append(sample["rejected"])

        feature_dict = {
            "prompt": datasets.Value("string"),
            "rejected": datasets.Value("string"),
            "chosen": datasets.Value("string"),
        }

        ds = datasets.Dataset.from_dict(datasets_dict, features=datasets.Features(feature_dict))
        if train_size != 1:
            ds = ds.train_test_split(train_size=train_size, test_size=1 - train_size, seed=seed)

        return ds


class TrainingTaskForQuestionAnswering(BaseModel, TrainingData):
    """
    Training data for question answering

    Args:
        formatting_func: A formatting function converting a dictionary of records into zero,
            one or more question-context-answer text tuples.
        question: The TextField to use for training.
        context: The TextField to use for training.
        answer: The TextQuestion to use for training.

    Examples:
        >>> # with defaults
        >>> from argilla import TrainingTaskForQuestionAnswering
        >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
        >>> task = TrainingTaskForQuestionAnswering(
        ...     question=dataset.field_by_name("question"),
        ...     context=dataset.field_by_name("context"),
        ...     answer=dataset.question_by_name("answer"),
        ... )
        >>> dataset.prepare_for_training(framework="...", task=task)
        >>> # with formatting_func
        >>> from argilla import TrainingTaskForQuestionAnswering
        >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
        >>> def formatting_func(sample: Dict[str, Any]):
        ...     question = sample["question"]
        ...     context = sample["context"]
        ...     for answer in sample["answer"]:
        ...         if not all([question, context, answer["value"]]):
        ...             continue
        ...         yield question, context, answer["value"]
        >>> task = TrainingTaskForQuestionAnswering(formatting_func=formatting_func)
        >>> dataset.prepare_for_training(framework="...", task=task)
    """

    defaults: Optional[QuestionAnsweringDefaults] = None
    formatting_func: Optional[Callable[[Dict[str, Any]], Union[None, str, Iterator[str]]]] = None
    _formatting_func_return_types = QuestionAnsweringReturnTypes
    _supported_frameworks_names = ["transformers"]

    @property
    def question(self) -> TextField:
        return self.defaults.question

    @property
    def context(self) -> TextField:
        return self.defaults.context

    @property
    def answer(self) -> TextQuestion:
        return self.defaults.answer

    def _format_data(self, dataset: "FeedbackDataset") -> List[Dict[str, str]]:
        if self.formatting_func is not None:
            output = self._execute_formatting_func(dataset)
            return [
                {"question": question, "context": context, "answer": answer} for question, context, answer in output
            ]
        else:
            return super()._format_data(dataset)

    @requires_dependencies("transformers")
    def _prepare_for_training_with_transformers(
        self, data: List[dict], train_size: float, seed: int, framework=None
    ) -> Union["datasets.Dataset", "datasets.DatasetDict"]:
        import datasets

        datasets_dict = {
            "question": [],
            "context": [],
            "answer": [],
        }
        for entry in data:
            if any([entry["question"] is None, entry["context"] is None, entry["answer"] is None]):
                continue
            if entry["answer"] not in entry["context"]:
                warnings.warn("This is extractive QnA but the answer is not in the context.")
                continue
            # get index of answer in context
            answer_start = entry["context"].index(entry["answer"])
            datasets_dict["question"].append(entry["question"])
            datasets_dict["context"].append(entry["context"])
            datasets_dict["answer"].append({"answer_start": [answer_start], "text": [entry["answer"]]})

        datasets_dict["id"] = list(range(len(data)))
        feature_dict = {
            "question": datasets.Value("string"),
            "context": datasets.Value("string"),
            "answer": datasets.Sequence(
                feature={
                    "text": datasets.Value(dtype="string", id=None),
                    "answer_start": datasets.Value(dtype="int32", id=None),
                },
                length=-1,
                id=None,
            ),
            "id": datasets.Value(dtype="int32"),
        }

        ds = datasets.Dataset.from_dict(datasets_dict, features=datasets.Features(feature_dict))

        if train_size != 1:
            ds = ds.train_test_split(train_size=train_size, test_size=1 - train_size, seed=seed)

        return ds


class TrainingTaskForChatCompletion(BaseModel, TrainingData):
    """Training data for chat completion

    Args:
        formatting_func: A formatting function converting a dictionary of records into zero,
            one or more chat-turn-role-content text tuples.

    Examples:
        >>> from argilla import TrainingTaskForChatCompletion
        >>> dataset = rg.FeedbackDataset.from_argilla(name="...")
        >>> def formatting_func(sample: Dict[str, Any]):
        ...     from uuid import uuid4
        ...     chat_id = str(uuid4())
        ...     if sample["response"]:
        ...         chat = str(uuid4())
        ...         user_message = sample["prompt"][0]["value"]
        ...         system_message = sample["response"][0]["value"]
        ...         yield [(chat, "0", "user", user_message), (chat, "1", "assistant", system_message)]
        >>> task = TrainingTaskForChatCompletion(formatting_func=formatting_func)
        >>> dataset.prepare_for_training(framework="...", task=task)
    """

    formatting_func: Callable[[Dict[str, Any]], Union[None, Dict[str, str], Iterator[Dict[str, str]]]]
    _formatting_func_return_types = ChatCompletionReturnTypes
    _supported_frameworks_names = ["openai"]

    def _format_data(self, dataset: "FeedbackDataset") -> List[Dict[str, str]]:
        output = self._execute_formatting_func(dataset)
        return [{"chat": chat, "turn": turn, "role": role, "content": content} for chat, turn, role, content in output]

    @requires_dependencies("openai>=0.27.10")
    def _prepare_for_training_with_openai(self, data: List[dict], train_size: float, seed: int) -> List[dict]:
        import datasets

        def _dict_to_format(ds: datasets.Dataset) -> List[Dict[str, List[Dict[str, str]]]]:
            """OpenAI expects a list of chats, each chat is a dict with a list of messages.
            Each message {"role": "user", "content": "Hello!"}
            """
            chats = []
            df = ds.to_pandas()
            for chat_id in df["chat"].unique():
                df_filter = df[df["chat"] == chat_id]
                new_chat = {"messages": []}
                for entry in df_filter.to_dict(orient="records"):
                    new_chat["messages"].append({"role": entry["role"], "content": entry["content"]})
                chats.append(new_chat)

            return chats

        datasets_dict = {"chat": [], "turn": [], "role": [], "content": []}
        for entry in data:
            if entry["role"] not in ["system", "user", "assistant"]:
                raise ValueError("Role must be one of 'system', 'user', 'assistant'")
            datasets_dict["chat"].append(entry["chat"])
            datasets_dict["turn"].append(entry["turn"])
            datasets_dict["role"].append(entry["role"])
            datasets_dict["content"].append(entry["content"])

        feature_dict = {
            "chat": datasets.Value("string"),
            "turn": datasets.Value("string"),
            "role": datasets.Value("string"),
            "content": datasets.Value("string"),
        }

        ds = datasets.Dataset.from_dict(datasets_dict, features=datasets.Features(feature_dict))
        ds = ds.sort(column_names=["chat", "turn"])

        if train_size != 1:
            ds = ds.train_test_split(
                train_size=train_size, test_size=1 - train_size, seed=seed, stratify_by_column="chat"
            )
            return _dict_to_format(ds["train"]), _dict_to_format(ds["test"])
        else:
            return _dict_to_format(ds)


class TrainingTaskForSentenceSimilarity(BaseModel, TrainingData):
    """Training data for sentence similarity.

    Args:
        formatting_func: A formatting function converting a dictionary of records into
            a dictionary of a pair of sentences, a pair of sentences and a label,
            a sentence and a label or a triplet of sentences.

    Examples:
        Example for argilla/emotion dataset:
        >>> from argilla import TrainingTaskForSentenceSimilarity
        >>> dataset = rg.FeedbackDataset.from_argilla(name="argilla/emotion")
        >>> def formatting_func(sample: Dict[str, Any]):
        ...     return {"sentence": sample["text"], "label": int(sample["label"][0]["value"])}
        >>> task = TrainingTaskForSentenceSimilarity(formatting_func=formatting_func)
        >>> dataset.prepare_for_training(framework="...", task=task)
    """

    defaults: Optional[SentenceSimilarityDefaults] = SentenceSimilarityDefaults()
    formatting_func: Callable[
        [Dict[str, Any]],
        Union[
            None, Dict[str, Union[float, int]], Dict[str, str], List[Dict[str, Union[float, int]]], List[Dict[str, str]]
        ],
    ] = None
    _formatting_func_return_types = SentenceSimilarityReturnTypes
    _supported_frameworks_names = ["sentence-transformers"]

    @property
    def __all_labels__(self) -> Optional[List[str]]:
        if self.label:
            return self.label.question.__all_labels__

    @property
    def __label2id__(self) -> Optional[Dict[str, int]]:
        if self.label:
            return self.label.question.__label2id__

    @property
    def __id2label__(self) -> Optional[Dict[int, str]]:
        if self.label:
            return self.label.question.__id2label__

    @property
    def label(
        self,
    ) -> Optional[Union[LabelQuestion, RatingQuestion, LabelQuestionUnification, RankingQuestionUnification]]:
        return self.defaults.label

    @property
    def texts(self) -> Optional[List[str]]:
        return self.defaults.texts

    def _format_data(self, dataset: "FeedbackDataset") -> List[Dict[str, Any]]:
        if self.formatting_func:
            output = self._execute_formatting_func(dataset)

            if "label" in output[0]:
                _all_labels = set()
                for sample in output:
                    if isinstance(sample, (list, tuple, set)):
                        for response in sample:
                            _all_labels.add(response["label"])
                    else:
                        _all_labels.add(sample["label"])

                if self.defaults.label is None:
                    labels = list(_all_labels)
                    if isinstance(labels[0], int):
                        label = RatingQuestionUnification(
                            question=RatingQuestion(name="custom_func", values=labels), strategy="majority"
                        )
                    else:
                        label = LabelQuestionUnification(question=LabelQuestion(name="custom_func", labels=labels))
                    self.defaults.label = label

            return output

        else:
            formatted_data = super()._format_data(dataset)
            # NOTE: Maybe this post processing of the formatted data can be simplified
            # or directly done in super()._format_data(dataset).
            new_keys = {field.name: f"sentence-{i}" for i, field in enumerate(self.texts, start=1)}
            if self.defaults.label:
                new_keys.update({self.defaults.label.question.name: "label"})

            outputs = []
            for example in formatted_data:
                record = {}
                for k, v in new_keys.items():
                    if v == "label":
                        value = example[v]
                        # At this point the label must be either an int or a float, determine which one is it.
                        if isinstance(value, str):
                            if value.lstrip("-").isdigit():
                                value = int(value)
                            else:
                                value = float(value)
                    else:
                        value = example[k]

                    record[v] = value
                outputs.append(record)

            return outputs

    def compute_unified_responses(self, responses: List[FeedbackRecord]):
        self.label.strategy.compute_unified_responses(responses=responses, field=self.label.question)

    @requires_dependencies("scikit-learn")
    def _train_test_split(
        self, data: List[dict], train_size: float, seed: int, stratify=None
    ) -> Tuple[List[dict], List[dict]]:
        from sklearn.model_selection import train_test_split

        return train_test_split(data, train_size=train_size, shuffle=True, random_state=seed, stratify=stratify)

    @requires_dependencies("sentence-transformers")
    def _prepare_for_training_with_sentence_transformers(
        self, data: Union[List[dict], List[List[dict]]], train_size: float, seed: int
    ) -> Union["InputExample", Tuple["InputExample", "InputExample"]]:
        from sentence_transformers import InputExample

        if not len(data) > 0:
            raise ValueError("The dataset must contain at least one sample to be able to train.")

        # Use the first sample to decide what type of dataset to generate:
        if isinstance(data[0], list):
            # In case we are returning lists, extract the first element of that list to check the fields.
            sample_keys = set(data[0][0].keys())
        elif isinstance(data[0], dict):
            sample_keys = set(data[0].keys())
        else:
            raise ValueError(f"The type is not supported: {type(data[0])}.")

        if sample_keys == {"label", "sentence-1", "sentence-2"}:

            def dataset_fields(sample):
                return {"texts": [sample["sentence-1"], sample["sentence-2"]], "label": sample["label"]}

        elif sample_keys == {"label", "sentence-1", "sentence-2", "sentence-3"}:

            def dataset_fields(sample):
                return {
                    "texts": [sample["sentence-1"], sample["sentence-2"], sample["sentence-3"]],
                    "label": sample["label"],
                }

        elif sample_keys == {"sentence-1", "sentence-2"}:

            def dataset_fields(sample):
                return {"texts": [sample["sentence-1"], sample["sentence-2"]]}

        elif sample_keys == {"sentence-1", "sentence-2", "sentence-3"}:

            def dataset_fields(sample):
                return {"texts": [sample["sentence-1"], sample["sentence-2"], sample["sentence-3"]]}

        elif sample_keys == {"label", "sentence"}:
            raise ValueError(
                "Datasets containing a `sentence` and a `label` should be transformed "
                "to contain triplets of `sentence-1`, `sentence-2`, `sentence-3` and `label`."
                r"An example can be seen at: https://github.com/UKPLab/sentence-transformers/blob/master/examples/training/other/training_batch_hard_trec.py"
            )
        else:
            raise ValueError(
                "Labeled datasets must contain a pair of `sentence-1` and "
                "`sentence-2` or triplets `sentence-1`, `sentence-2`, `sentence-3` "
                "and an optional `label`."
            )

        train_samples = []
        for sample in data:
            if isinstance(sample, list):
                for record in sample:
                    train_samples.append(InputExample(**dataset_fields(record)))
            else:
                train_samples.append(InputExample(**dataset_fields(sample)))

        if train_size != 1:
            stratify = None
            if (label := train_samples[0].label) and isinstance(label, int):
                stratify = [example.label for example in train_samples]

            train_data, test_data = self._train_test_split(train_samples, train_size, seed, stratify=stratify)
            return train_data, test_data
        else:
            return train_samples


TrainingTaskTypes = Union[
    TrainingTaskForTextClassification,
    TrainingTaskForSFT,
    TrainingTaskForRM,
    TrainingTaskForPPO,
    TrainingTaskForDPO,
    TrainingTaskForChatCompletion,
    TrainingTaskForSentenceSimilarity,
]


# Helper map fr the creation of the model cards.
TRAINING_TASK_MAPPING = {
    TrainingTaskForTextClassification: "for_text_classification",
    TrainingTaskForSFT: "for_supervised_fine_tuning",
    TrainingTaskForRM: "for_reward_modeling",
    TrainingTaskForPPO: "for_proximal_policy_optimization",
    TrainingTaskForDPO: "for_direct_preference_optimization",
    TrainingTaskForChatCompletion: "for_chat_completion",
    TrainingTaskForQuestionAnswering: "for_question_answering",
    TrainingTaskForSentenceSimilarity: "for_sentence_similarity",
}


# Old, deprecated variants.
class RenamedDeprecationMixin:
    @classmethod
    def warn(cls) -> None:
        this_class_name = cls.__name__
        first_subclass_name = cls.__mro__[1].__name__
        warnings.warn(
            (f"`{this_class_name}` has been renamed to `{first_subclass_name}`, please use the latter."),
            DeprecationWarning,
            stacklevel=3,
        )


class TrainingTaskMapping(TrainingTask, RenamedDeprecationMixin):
    @classmethod
    def for_text_classification(cls, *args, **kwargs) -> TrainingTaskForTextClassification:
        cls.warn()
        return super().for_text_classification(*args, **kwargs)

    @classmethod
    def for_supervised_fine_tuning(cls, *args, **kwargs) -> TrainingTaskForSFT:
        cls.warn()
        return super().for_supervised_fine_tuning(*args, **kwargs)

    @classmethod
    def for_reward_modeling(cls, *args, **kwargs) -> TrainingTaskForRM:
        cls.warn()
        return super().for_reward_modeling(*args, **kwargs)

    @classmethod
    def for_proximal_policy_optimization(cls, *args, **kwargs) -> TrainingTaskForPPO:
        cls.warn()
        return super().for_proximal_policy_optimization(cls, *args, **kwargs)

    @classmethod
    def for_direct_preference_optimization(cls, *args, **kwargs) -> TrainingTaskForDPO:
        cls.warn()
        return super().for_direct_preference_optimization(*args, **kwargs)

    @classmethod
    def for_chat_completion(cls, *args, **kwargs) -> TrainingTaskForChatCompletion:
        cls.warn()
        return super().for_chat_completion(*args, **kwargs)

    @classmethod
    def for_sentence_similarity(cls, *args, **kwargs) -> TrainingTaskForSentenceSimilarity:
        cls.warn()
        return super().for_sentence_similarity(*args, **kwargs)

    @classmethod
    def for_question_answering(cls, *args, **kwargs) -> TrainingTaskForQuestionAnswering:
        cls.warn()
        return super().for_question_answering(*args, **kwargs)


class TrainingTaskMappingForTextClassification(TrainingTaskForTextClassification, RenamedDeprecationMixin):
    def __init__(self, *args, **kwargs) -> None:
        self.warn()
        return super().__init__(*args, **kwargs)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

import typer

if TYPE_CHECKING:
    from argilla.client.workspaces import Workspace


def delete_user(
    ctx: typer.Context,
    username: str = typer.Argument(..., help="The username of the user to be removed from the workspace"),
) -> None:
    from argilla.cli.rich import echo_in_panel
    from argilla.cli.workspaces.utils import get_user

    workspace: "Workspace" = ctx.obj

    user = get_user(username)

    if user.is_owner:
        echo_in_panel(
            f"User with username={username} is an owner and cannot be removed from any workspace",
            title="User is owner",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1)

    try:
        workspace.delete_user(user.id)
    except ValueError as e:
        echo_in_panel(
            f"User with username={username} does not belong to the workspace={workspace.name}.",
            title="User not in workspace",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e
    except RuntimeError as e:
        echo_in_panel(
            "An unexpected error occurred when trying to delete the user from the workspace.",
            title="Unexpected error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e

    echo_in_panel(
        f"User with username={username} has been removed from workspace={workspace.name}",
        title="User deleted",
        title_align="left",
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
import warnings
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

from argilla.client.apis import AbstractApi, api_compatibility
from argilla.client.sdk.commons.errors import (
    AlreadyExistsApiError,
    ApiCompatibilityError,
    ForbiddenApiError,
    NotFoundApiError,
)
from argilla.client.sdk.datasets.api import get_dataset
from argilla.client.sdk.datasets.models import TaskType
from argilla.pydantic_v1 import BaseModel, Field


@dataclass
class _AbstractSettings:
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "_AbstractSettings":
        """
        Creates a settings instance from a plain python dictionary

        Args:
            data: The data dict

        Returns:
            A new ``cls`` settings instance
        """
        raise NotImplementedError()


@dataclass
class LabelsSchemaSettings(_AbstractSettings):
    """
    A base dataset settings class for labels schema management

    Args:
        label_schema: The label's schema for the dataset

    """

    label_schema: List[str]

    def __post_init__(self):
        if not isinstance(self.label_schema, (set, list, tuple)):
            raise ValueError(
                f"`label_schema` is of type={type(self.label_schema)}, but type=set is preferred, and also both type=list and type=tuple are allowed."
            )
        self.label_schema = self._get_unique_labels()

    def _get_unique_labels(self) -> List[str]:
        unique_labels = []
        for label in self.label_schema:
            if label not in unique_labels:
                unique_labels.append(label)
        return unique_labels

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LabelsSchemaSettings":
        label_schema = data.get("label_schema", {})
        labels = [label["name"] for label in label_schema.get("labels", [])]
        return cls(label_schema=labels)

    @property
    def label2id(self) -> Dict[str, int]:
        return {label: i for i, label in enumerate(self.label_schema)}

    @property
    def id2label(self) -> Dict[int, str]:
        return {i: label for i, label in enumerate(self.label_schema)}


@dataclass
class TextClassificationSettings(LabelsSchemaSettings):
    """
    The settings for text classification datasets

    Args:
        label_schema: The label's schema for the dataset

    """


@dataclass
class TokenClassificationSettings(LabelsSchemaSettings):
    """
    The settings for token classification datasets

    Args:
        label_schema: The label's schema for the dataset

    """


Settings = Union[TextClassificationSettings, TokenClassificationSettings]

__TASK_TO_SETTINGS__ = {
    TaskType.text_classification: TextClassificationSettings,
    TaskType.token_classification: TokenClassificationSettings,
}


class Datasets(AbstractApi):
    """Dataset client api class"""

    _API_PREFIX = "/api/datasets"

    class _DatasetApiModel(BaseModel):
        id: Optional[str]
        name: str
        task: TaskType
        owner: Optional[str] = None
        workspace: Optional[str] = None
        created_at: Optional[datetime] = None
        last_updated: Optional[datetime] = None

        tags: Dict[str, str] = Field(default_factory=dict)
        metadata: Dict[str, Any] = Field(default_factory=dict)

    class _SettingsApiModel(BaseModel):
        label_schema: Dict[str, Any]

    def find_by_name(self, name: str, workspace: Optional[str] = None) -> _DatasetApiModel:
        dataset = get_dataset(self.http_client, name=name, workspace=workspace).parsed
        return self._DatasetApiModel.parse_obj(dataset)

    def create(self, name: str, task: TaskType, workspace: str) -> _DatasetApiModel:
        try:
            with api_compatibility(self, min_version="1.4.0"):
                dataset = self._DatasetApiModel(name=name, task=task, workspace=workspace)
                self.http_client.post(f"{self._API_PREFIX}", json=dataset.dict())
        except ApiCompatibilityError:
            dataset = self._DatasetApiModel(name=name, task=task)
            self.http_client.post(f"{self._API_PREFIX}?workspace={workspace}", json=dataset.dict())
        return dataset

    def configure(self, name: str, workspace: str, settings: Settings):
        """
        Configures dataset settings. If dataset does not exist, a new one will be created.
        Pass only settings that want to configure

        Args:
            name: The dataset name
            workspace: The dataset workspace
            settings: The dataset settings
        """
        try:
            task = (
                TaskType.text_classification
                if isinstance(settings, TextClassificationSettings)
                else TaskType.token_classification
            )
            ds = self.create(name=name, task=task, workspace=workspace)
        except AlreadyExistsApiError:
            ds = self.find_by_name(name, workspace=workspace)
        self._save_settings(dataset=ds, settings=settings)

    def scan(
        self,
        name: str,
        projection: Optional[Set[str]] = None,
        limit: Optional[int] = None,
        sort: Optional[List[Tuple[str, str]]] = None,
        id_from: Optional[str] = None,
        batch_size: int = 250,
        **query,
    ) -> Iterable[dict]:
        """
        Search records over a dataset

        Args:
            name: the dataset
            query: the search query
            projection: a subset of record fields to retrieve. If not provided,
                 only id's will be returned
            sort: The fields on which to sort [(<field_name>, 'asc|decs')].
            limit: The number of records to retrieve
            id_from: If provided, starts gathering the records starting from that Record.
                As the Records returned with the load method are sorted by ID, ´id_from´
                can be used to load using batches.
            batch_size: If provided, load `batch_size` samples per request. A lower batch
                size may help avoid timeouts.

        Returns:
            An iterable of raw object containing per-record info
        """

        if limit and limit < 0:
            raise ValueError("The scan limit must be non-negative.")

        limit = limit if limit else math.inf
        url = f"{self._API_PREFIX}/{name}/records/:search?limit={{limit}}"
        query = self._parse_query(query=query)

        request = {
            "fields": list(projection) if projection else ["id"],
            "query": query,
        }

        if sort is not None:
            try:
                if isinstance(sort, list):
                    assert all([(isinstance(item, tuple)) and (item[-1] in ["asc", "desc"]) for item in sort])
                else:
                    raise Exception()
            except Exception:
                raise ValueError("sort must be a dict formatted as List[Tuple[<field_name>, 'asc|desc']]")
            request["sort_by"] = [{"id": item[0], "order": item[-1]} for item in sort]

        elif id_from:
            # TODO: Show message since sort + next_id is not compatible since fixes a sort by id
            request["next_idx"] = id_from

        with api_compatibility(self, min_version="1.2.0"):
            request_limit = min(limit, batch_size)
            response = self.http_client.post(
                url.format(limit=request_limit),
                json=request,
            )

            while response.get("records"):
                yield from response["records"]
                limit -= request_limit
                if limit <= 0:
                    return

                next_request_params = {k: response[k] for k in ["next_idx", "next_page_cfg"] if response.get(k)}
                if not next_request_params:
                    return

                request_limit = min(limit, batch_size)
                response = self.http_client.post(
                    path=url.format(limit=request_limit),
                    json={**request, **next_request_params},
                )

    def update_record(
        self,
        name: str,
        record_id: str,
        **content,
    ):
        with api_compatibility(self, min_version="1.2.0"):
            url = f"{self._API_PREFIX}/{name}/records/{record_id}"
            response = self.http_client.patch(
                path=url,
                json=content,
            )
            return response

    def delete_records(
        self,
        name: str,
        mark_as_discarded: bool = False,
        discard_when_forbidden: bool = True,
        **query: Optional[dict],
    ) -> Tuple[int, int]:
        """
        Tries to delete records in a dataset for a given query/ids list.

        Args:
            name: The dataset name
            query: The query matching records
            ids: A list of records ids. If provided, the query param will be ignored
            mark_as_discarded: If `True`, the matched records will be marked as `Discarded` instead
                of delete them
            discard_when_forbidden: Only super-user or dataset creator can delete records from a dataset.
                So, running "hard" deletion for other users will raise an `ForbiddenApiError` error.
                If this parameter is `True`, the client API will automatically try to mark as ``Discarded``
                records instead.

        Returns:
            The total of matched records and real number of processed errors. These numbers could not
            be the same if some data conflicts are found during operations (some matched records change during
            deletion).

        """
        with api_compatibility(self, min_version="0.18"):
            try:
                query = self._parse_query(query=query)
                response = self.http_client.delete(
                    path=f"{self._API_PREFIX}/{name}/data?mark_as_discarded={mark_as_discarded}",
                    json=query,
                )
                return response["matched"], response["processed"]
            except ForbiddenApiError as faer:
                if discard_when_forbidden:
                    warnings.warn(
                        message=f"{faer}. Records will be discarded instead",
                        category=UserWarning,
                    )
                    return self.delete_records(
                        name=name,
                        query=query,
                        mark_as_discarded=True,
                        discard_when_forbidden=False,  # Next time will raise the error
                    )
                else:
                    raise faer

    def _save_settings(self, dataset: _DatasetApiModel, settings: Settings):
        if __TASK_TO_SETTINGS__.get(dataset.task) != type(settings):
            raise ValueError(
                f"The provided settings type {type(settings)} cannot be applied to dataset. Task type mismatch"
            )

        settings_ = self._SettingsApiModel.parse_obj(
            {"label_schema": {"labels": [label for label in settings.label_schema]}}
        )

        try:
            with api_compatibility(self, min_version="1.4"):
                self.http_client.patch(
                    f"{self._API_PREFIX}/{dataset.name}/{dataset.task.value}/settings?workspace={dataset.workspace}",
                    json=settings_.dict(),
                )
        except ApiCompatibilityError:
            with api_compatibility(self, min_version="0.15"):
                self.http_client.put(
                    f"{self._API_PREFIX}/{dataset.task.value}/{dataset.name}/settings",
                    json=settings_.dict(),
                )

    def load_settings(self, name: str, workspace: Optional[str] = None) -> Optional[Settings]:
        """
        Load the dataset settings

        Args:
            name: The dataset name
            workspace: The workspace name where the dataset belongs to

        Returns:
            Settings defined for the dataset
        """
        dataset = self.find_by_name(name, workspace=workspace)
        try:
            with api_compatibility(self, min_version="1.0"):
                params = {"workspace": dataset.workspace} if dataset.workspace else {}
                response = self.http_client.get(
                    f"{self._API_PREFIX}/{dataset.name}/{dataset.task.value}/settings", params=params
                )
                return __TASK_TO_SETTINGS__.get(dataset.task).from_dict(response)
        except NotFoundApiError:
            return None
        except Exception:
            return None



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Dict, List, Optional, Union

import argilla as rg
from argilla.client import singleton
from argilla.client.api import load
from argilla.client.models import TextClassificationRecord
from argilla.client.sdk.text_classification.models import LabelingRule


class Rule:
    """A rule (labeling function) in form of an ElasticSearch query.

    Args:
        query: An ElasticSearch query with the `query string syntax <https://docs.argilla.io/en/latest/practical_guides/filter_dataset.html>`_.
        label: The label associated to the query. Can also be a list of labels.
        name: An optional name for the rule to be used as identifier in the
            `argilla.labeling.text_classification.WeakLabels` class. By default, we will use the ``query`` string.

    Examples:
        >>> import argilla as rg
        >>> urgent_rule = Rule(query="inputs.text:(urgent AND immediately)", label="urgent", name="urgent_rule")
        >>> not_urgent_rule = Rule(query="inputs.text:(NOT urgent) AND metadata.title_length>20", label="not urgent")
        >>> not_urgent_rule.apply("my_dataset")
        >>> my_dataset_records = rg.load(name="my_dataset")
        >>> not_urgent_rule(my_dataset_records[0])
        "not urgent"
    """

    def __init__(
        self,
        query: str,
        label: Union[str, List[str]],
        name: Optional[str] = None,
        author: Optional[str] = None,
    ):
        self._query = query
        self._label = label
        self._name = name
        self._author = author
        self._matching_ids = None

    @property
    def query(self) -> str:
        """The rule query"""
        return self._query

    @property
    def label(self) -> Union[str, List[str]]:
        """The rule label"""
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def name(self):
        """The name of the rule."""
        if self._name is not None:
            return self._name
        return self._query

    @property
    def author(self):
        """Who authored the rule."""
        return self._author

    def _convert_to_labeling_rule(self):
        """Converts the rule to a LabelingRule"""
        if isinstance(self._label, str):
            labels = [self._label]
        else:
            labels = self._label

        return LabelingRule(query=self.query, labels=labels)

    def add_to_dataset(self, dataset: str):
        """Add to rule to the given dataset"""
        singleton.active_api().add_dataset_labeling_rules(dataset, rules=[self._convert_to_labeling_rule()])

    def remove_from_dataset(self, dataset: str):
        """Removes the rule from the given dataset"""

        singleton.active_api().delete_dataset_labeling_rules(dataset, rules=[self._convert_to_labeling_rule()])

    def update_at_dataset(self, dataset: str):
        """Updates the rule at the given dataset"""
        singleton.active_api().update_dataset_labeling_rules(dataset, rules=[self._convert_to_labeling_rule()])

    def apply(self, dataset: str):
        """Apply the rule to a dataset and save matching ids of the records.

        Args:
            dataset: The name of the dataset.
        """
        records = load(name=dataset, query=self._query, include_vectors=False)

        self._matching_ids = {record.id: None for record in records}

    def metrics(self, dataset: str) -> Dict[str, Union[int, float]]:
        """Compute the rule metrics for a given dataset:

        - **coverage**: Fraction of the records labeled by the rule.
        - **annotated_coverage**: Fraction of annotated records labeled by the rule.
        - **correct**: Number of records the rule labeled correctly (if annotations are available).
        - **incorrect**: Number of records the rule labeled incorrectly (if annotations are available).
        - **precision**: Fraction of correct labels given by the rule (if annotations are available). The precision does not penalize the rule for abstains.

        Args:
            dataset: Name of the dataset for which to compute the rule metrics.

        Returns:
            The rule metrics.
        """
        metrics = singleton.active_api().rule_metrics_for_dataset(
            dataset=dataset,
            rule=LabelingRule(query=self.query, label=self.label),
        )

        return {
            "coverage": metrics.coverage,
            "annotated_coverage": metrics.coverage_annotated,
            "correct": int(metrics.correct) if metrics.correct is not None else None,
            "incorrect": int(metrics.incorrect) if metrics.incorrect is not None else None,
            "precision": metrics.precision if metrics.precision is not None else None,
        }

    def __call__(self, record: TextClassificationRecord) -> Optional[Union[str, List[str]]]:
        """Check if the given record is among the matching ids from the ``self.apply`` call.

        Args:
            record: The record to be labelled.

        Returns:
            A label or list of labels if the record id is among the matching ids, otherwise None.

        Raises:
            RuleNotAppliedError: If the rule was not applied to the dataset before.
        """
        if self._matching_ids is None:
            raise RuleNotAppliedError("Rule was still not applied. Please call `self.apply(dataset)` first.")

        try:
            self._matching_ids[record.id]
        except KeyError:
            return None
        else:
            return self._label

    def __repr__(self):
        """The rule representation."""
        return f"Rule(query='{self.query}', label='{self.label}', name='{self.name}')"

    def __str__(self):
        """The rule string representation."""
        return repr(self)


def add_rules(dataset: str, rules: List[Rule]):
    """Adds the rules to a given dataset

    Args:
        dataset: Name of the dataset.
        rules: Rules to add to the dataset

    Returns:
    """
    rules = [rule._convert_to_labeling_rule() for rule in rules]
    return singleton.active_api().add_dataset_labeling_rules(dataset, rules)


def delete_rules(dataset: str, rules: List[Rule]):
    """Deletes the rules from the given dataset

    Args:
        dataset: Name of the dataset
        rules: Rules to delete from the dataset

    Returns:
    """
    rules = [rule._convert_to_labeling_rule() for rule in rules]
    singleton.active_api().delete_dataset_labeling_rules(dataset, rules)


def update_rules(dataset: str, rules: List[Rule]):
    """Updates the rules of the given dataset

    Args:
        dataset: Name of the dataset
        rules: Rules to update at the dataset

    Returns:
    """
    rules = [rule._convert_to_labeling_rule() for rule in rules]
    singleton.active_api().update_dataset_labeling_rules(dataset, rules)


def load_rules(dataset: str) -> List[Rule]:
    """load the rules defined in a given dataset.

    Args:
        dataset: Name of the dataset.

    Returns:
        A list of rules defined in the given dataset.
    """
    rules = singleton.active_api().fetch_dataset_labeling_rules(dataset)
    return [
        Rule(
            query=rule.query,
            label=rule.label or rule.labels,
            name=rule.description,
            author=rule.author,
        )
        for rule in rules
    ]


class RuleNotAppliedError(Exception):
    pass



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import textwrap
import warnings
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Tuple, Union

from argilla.client.feedback.constants import FETCHING_BATCH_SIZE
from argilla.client.feedback.dataset import helpers
from argilla.client.feedback.dataset.base import FeedbackDatasetBase, R
from argilla.client.feedback.dataset.local.mixins import ArgillaMixin, TaskTemplateMixin
from argilla.client.feedback.dataset.mixins import MetricsMixin, UnificationMixin
from argilla.client.feedback.integrations.huggingface.dataset import HuggingFaceDatasetMixin
from argilla.client.feedback.schemas.enums import RecordSortField, SortOrder
from argilla.client.feedback.schemas.records import FeedbackRecord
from argilla.client.feedback.schemas.vector_settings import VectorSettings
from argilla.client.feedback.training.schemas.base import (
    TrainingTaskForChatCompletion,
    TrainingTaskForDPO,
    TrainingTaskForPPO,
    TrainingTaskForQuestionAnswering,
    TrainingTaskForRM,
    TrainingTaskForSentenceSimilarity,
    TrainingTaskForSFT,
    TrainingTaskForTextClassification,
    TrainingTaskTypes,
)
from argilla.client.models import Framework

if TYPE_CHECKING:
    from argilla.client.feedback.schemas.types import (
        AllowedFieldTypes,
        AllowedMetadataPropertyTypes,
        AllowedQuestionTypes,
    )


_LOGGER = logging.getLogger(__name__)


class FeedbackDataset(
    ArgillaMixin,
    HuggingFaceDatasetMixin,
    FeedbackDatasetBase[FeedbackRecord],
    TaskTemplateMixin,
    MetricsMixin,
    UnificationMixin,
):
    def __init__(
        self,
        *,
        fields: List["AllowedFieldTypes"],
        questions: List["AllowedQuestionTypes"],
        metadata_properties: Optional[List["AllowedMetadataPropertyTypes"]] = None,
        vectors_settings: Optional[List[VectorSettings]] = None,
        guidelines: Optional[str] = None,
        allow_extra_metadata: bool = True,
    ) -> None:
        """Initializes a `FeedbackDataset` instance locally.

        Args:
            fields: contains the fields that will define the schema of the records in the dataset.
            questions: contains the questions that will be used to annotate the dataset.
            metadata_properties: contains the metadata properties that will be indexed
                and could be used to filter the dataset. Defaults to `None`.
            vectors_settings: contains the vectors settings that will define the configuration
                of the vectors associated to the records in the dataset and that would
                allow to perform vector search. Defaults to `None`.
            guidelines: contains the guidelines for annotating the dataset. Defaults to `None`.
            allow_extra_metadata: whether to allow extra metadata that has not been defined
                as a metadata property in the records. Defaults to `True`.

        Raises:
            TypeError: if `fields` is not a list of `FieldSchema`.
            ValueError: if `fields` does not contain at least one required field.
            TypeError: if `questions` is not a list of `TextQuestion`, `RatingQuestion`,
                `LabelQuestion`, and/or `MultiLabelQuestion`.
            ValueError: if `questions` does not contain at least one required question.
            TypeError: if `guidelines` is not None and not a string.
            ValueError: if `guidelines` is an empty string.

        Examples:
            >>> import argilla as rg
            >>> rg.init(api_url="...", api_key="...")
            >>> dataset = rg.FeedbackDataset(
            ...     fields=[
            ...         rg.TextField(name="text", required=True),
            ...         rg.TextField(name="label", required=True),
            ...     ],
            ...     questions=[
            ...         rg.TextQuestion(
            ...             name="question-1",
            ...             description="This is the first question",
            ...             required=True,
            ...         ),
            ...         rg.RatingQuestion(
            ...             name="question-2",
            ...             description="This is the second question",
            ...             required=True,
            ...             values=[1, 2, 3, 4, 5],
            ...         ),
            ...         rg.LabelQuestion(
            ...             name="question-3",
            ...             description="This is the third question",
            ...             required=True,
            ...             labels=["positive", "negative"],
            ...         ),
            ...         rg.MultiLabelQuestion(
            ...             name="question-4",
            ...             description="This is the fourth question",
            ...             required=True,
            ...             labels=["category-1", "category-2", "category-3"],
            ...         ),
            ...     ],
            ...     metadata_properties=[
            ...         rg.TermsMetadataProperty(
            ...             name="metadata-property-1",
            ...             values=["a", "b", "c"]
            ...         ),
            ...         rg.IntegerMetadataProperty(
            ...             name="metadata-property-2",
            ...             gt=0,
            ...             lt=10,
            ...         ),
            ...         rg.FloatMetadataProperty(
            ...             name="metadata-property-2",
            ...             gt=-10.0,
            ...             lt=10.0,
            ...         ),
            ...     ],
            ...     guidelines="These are the annotation guidelines.",
            ... )
        """

        helpers.validate_fields(fields)
        helpers.validate_questions(questions)
        helpers.validate_metadata_properties(metadata_properties)

        if guidelines is not None:
            if not isinstance(guidelines, str):
                raise TypeError(
                    f"Expected `guidelines` to be either None (default) or a string, got {type(guidelines)} instead."
                )
            if len(guidelines) < 1:
                raise ValueError(
                    "Expected `guidelines` to be either None (default) or a non-empty string, minimum length is 1."
                )

        self._fields = fields or []
        self._questions = questions or []
        self._metadata_properties = metadata_properties or []
        self._guidelines = guidelines
        self._allow_extra_metadata = allow_extra_metadata

        if vectors_settings:
            self._vectors_settings = {vector_setting.name: vector_setting for vector_setting in vectors_settings}
        else:
            self._vectors_settings: Dict[str, VectorSettings] = {}
        self._records = []

    @property
    def guidelines(self) -> Optional[str]:
        return self._guidelines

    @property
    def allow_extra_metadata(self) -> bool:
        return self._allow_extra_metadata

    @property
    def fields(self) -> Union[List["AllowedFieldTypes"]]:
        return self._fields

    @property
    def questions(self) -> Union[List["AllowedQuestionTypes"]]:
        return self._questions

    @property
    def metadata_properties(
        self,
    ) -> Union[List["AllowedMetadataPropertyTypes"]]:
        return self._metadata_properties

    @property
    def vectors_settings(self) -> List["VectorSettings"]:
        """Returns the vector settings of the dataset."""
        return [v for v in self._vectors_settings.values()]

    @property
    def records(self) -> List["FeedbackRecord"]:
        """Returns the records in the dataset."""
        return self._records

    def __repr__(self) -> str:
        """Returns a string representation of the dataset."""
        indent = "   "
        return (
            "FeedbackDataset("
            + textwrap.indent(f"\nfields={self.fields}", indent)
            + textwrap.indent(f"\nquestions={self.questions}", indent)
            + textwrap.indent(f"\nguidelines={self.guidelines})", indent)
            + textwrap.indent(f"\nmetadata_properties={self.metadata_properties})", indent)
            + textwrap.indent(f"\nvectors_settings={self.vectors_settings})", indent)
            + "\n)"
        )

    def __len__(self) -> int:
        """Returns the number of records in the dataset."""
        return len(self._records)

    def __getitem__(self, key: Union[slice, int]) -> Union["FeedbackRecord", List["FeedbackRecord"]]:
        """Returns the record(s) at the given index(es).

        Args:
            key: the index(es) of the record(s) to return. Can either be a single index or a slice.

        Returns:
            Either the record of the given index, or a list with the records at the given indexes.
        """
        if len(self._records) < 1:
            raise RuntimeError(
                "In order to get items from `FeedbackDataset` you need to add them first" " with `add_records`."
            )
        if isinstance(key, int) and len(self._records) < key:
            raise IndexError(f"This dataset contains {len(self)} records, so index {key} is out of range.")
        return self._records[key]

    def iter(self, batch_size: Optional[int] = FETCHING_BATCH_SIZE) -> Iterator[List["FeedbackRecord"]]:
        """Returns an iterator over the records in the dataset.

        Args:
            batch_size: the size of the batches to return. Defaults to 100.
        """
        for i in range(0, len(self._records), batch_size):
            yield self._records[i : i + batch_size]

    def add_records(
        self, records: Union["FeedbackRecord", Dict[str, Any], List[Union["FeedbackRecord", Dict[str, Any]]]]
    ) -> None:
        """Adds the given records to the dataset, and stores them locally. If you are
        planning to push those to Argilla, you will need to call `push_to_argilla` afterwards,
        to both create the dataset in Argilla and push the records to it. Then, from a
        `FeedbackDataset` pushed to Argilla, you'll just need to call `add_records` and
        those will be automatically uploaded to Argilla.

        Args:
            records: can be a single `FeedbackRecord`, a list of `FeedbackRecord`,
                a single dictionary, or a list of dictionaries. If a dictionary is provided,
                it will be converted to a `FeedbackRecord` internally.

        Raises:
            ValueError: if the given records are an empty list.
            ValueError: if the given records are neither: `FeedbackRecord`, list of `FeedbackRecord`,
                list of dictionaries as a record or dictionary as a record.
            ValueError: if the given records do not match the expected schema.
        """
        records = helpers.normalize_records(records)
        helpers.validate_dataset_records(self, records)

        if len(self._records) > 0:
            self._records += records
        else:
            self._records = records

    def add_metadata_property(
        self, metadata_property: "AllowedMetadataPropertyTypes"
    ) -> "AllowedMetadataPropertyTypes":
        """Adds the given metadata property to the dataset.

        Args:
            metadata_property: the metadata property to add.

        Returns:
            The metadata property that was added.

        Raises:
            TypeError: if `metadata_property` is not a `MetadataPropertySchema`.
            ValueError: if `metadata_property` is already in the dataset.
        """
        self._unique_metadata_property(metadata_property)
        self._metadata_properties.append(metadata_property)
        return metadata_property

    def add_vector_settings(self, vector_settings: VectorSettings) -> VectorSettings:
        if self._vectors_settings.get(vector_settings.name):
            raise ValueError(f"Vector settings with name '{vector_settings.name}' already exists in the dataset.")

        self._vectors_settings[vector_settings.name] = vector_settings
        return vector_settings

    def update_vectors_settings(self, vectors_settings: Union[VectorSettings, List[VectorSettings]]) -> None:
        """Does nothing because the `vector_settings` are updated automatically for
        `FeedbackDataset` datasets when assigning their updateable attributes to a new value.
        """
        warnings.warn(
            "`update_vectors_settings` method is not supported for `FeedbackDataset` datasets"
            " unless its pushed to Argilla i.e. `RemoteFeedbackDataset`. This is because the"
            " `vector_settings` updates are already applied via assignment if any. So,"
            " this method is not required locally.",
            UserWarning,
            stacklevel=1,
        )

    def delete_vectors_settings(
        self, vectors_settings: Union[str, List[str]]
    ) -> Union[VectorSettings, List[VectorSettings]]:
        """Deletes the given vector settings from the dataset.

        Args:
            vectors_settings: the name/s of the vector settings to delete.

        Returns:
            The vector settings that were deleted.

        Raises:
            ValueError: if the provided `vectors_settings` is/are not in the dataset.
        """
        if isinstance(vectors_settings, str):
            vectors_settings = [vectors_settings]

        if not self.vectors_settings:
            raise ValueError(
                "The current `FeedbackDataset` does not contain any `vectors_settings` defined, so"
                " none can be deleted."
            )

        if not all(vector_setting in self._vectors_settings.keys() for vector_setting in vectors_settings):
            raise ValueError(
                f"Invalid `vectors_settings={vectors_settings}` provided. It cannot be"
                " deleted because it does not exist, make sure you delete just existing `vectors_settings`"
                " meaning that the name matches any of the existing `vectors_settings` if any. Current"
                f" `vectors_settings` are: '{', '.join(self._vectors_settings.keys())}'."
            )

        deleted_vectors_settings = []
        for vector_setting in vectors_settings:
            deleted_vectors_settings.append(self._vectors_settings.pop(vector_setting))
        return deleted_vectors_settings if len(deleted_vectors_settings) > 1 else deleted_vectors_settings[0]

    def update_metadata_properties(
        self,
        metadata_properties: Union["AllowedMetadataPropertyTypes", List["AllowedMetadataPropertyTypes"]],
    ) -> None:
        """Does nothing because the `metadata_properties` are updated automatically for
        `FeedbackDataset` datasets when assigning their updateable attributes to a new value.
        """
        warnings.warn(
            "`update_metadata_properties` method is not supported for `FeedbackDataset` datasets"
            " unless its pushed to Argilla i.e. `RemoteFeedbackDataset`. This is because the"
            " `metadata_properties` updates are already applied via assignment if any. So,"
            " this method is not required locally.",
            UserWarning,
            stacklevel=1,
        )

    def delete_metadata_properties(
        self, metadata_properties: Union[str, List[str]]
    ) -> Union["AllowedMetadataPropertyTypes", List["AllowedMetadataPropertyTypes"]]:
        """Deletes the given metadata properties from the dataset.

        Args:
            metadata_properties: the name/s of the metadata property/ies to delete.

        Returns:
            The metadata properties that were deleted.

        Raises:
            TypeError: if `metadata_properties` is not a string or a list of strings.
            ValueError: if the provided `metadata_properties` is/are not in the dataset.
        """
        if not isinstance(metadata_properties, list):
            metadata_properties = [metadata_properties]

        if not self.metadata_properties:
            raise ValueError(
                "The current `FeedbackDataset` does not contain any `metadata_properties` defined, so"
                " none can be deleted."
            )
        metadata_properties_mapping = {
            metadata_property.name: metadata_property for metadata_property in self.metadata_properties
        }
        if not all(
            metadata_property in metadata_properties_mapping.keys() for metadata_property in metadata_properties
        ):
            raise ValueError(
                f"Invalid `metadata_properties={metadata_properties}` provided. It cannot be"
                " deleted because it does not exist, make sure you delete just existing `metadata_properties`"
                " meaning that the name matches any of the existing `metadata_properties` if any. Current"
                f" `metadata_properties` are: '{', '.join(metadata_properties_mapping.keys())}'."
            )

        deleted_metadata_properties = []
        for metadata_property in metadata_properties:
            deleted_metadata_properties.append(metadata_properties_mapping.pop(metadata_property))
        self._metadata_properties = list(metadata_properties_mapping.values())
        return deleted_metadata_properties if len(deleted_metadata_properties) > 1 else deleted_metadata_properties[0]

    # TODO(alvarobartt,davidberenstein1957): we should consider having something like
    # `export(..., training=True)` to export the dataset records in any format, replacing
    # both `format_as` and `prepare_for_training`
    def prepare_for_training(
        self,
        framework: Union[Framework, str],
        task: TrainingTaskTypes,
        train_size: Optional[float] = 1,
        test_size: Optional[float] = None,
        seed: Optional[int] = None,
        lang: Optional[str] = None,
    ) -> Any:
        """
        Prepares the dataset for training for a specific training framework and NLP task by splitting the dataset into train and test sets.

        Args:
            framework: the framework to use for training. Currently supported frameworks are: `transformers`, `peft`,
                `setfit`, `spacy`, `spacy-transformers`, `span_marker`, `spark-nlp`, `openai`, `trl`, `sentence-transformers`.
            task: the NLP task to use for training. Currently supported tasks are: `TrainingTaskForTextClassification`,
                `TrainingTaskForSFT`, `TrainingTaskForRM`, `TrainingTaskForPPO`, `TrainingTaskForDPO`, `TrainingTaskForSentenceSimilarity`.
            train_size: the size of the train set. If `None`, the whole dataset will be used for training.
            test_size: the size of the test set. If `None`, the whole dataset will be used for testing.
            seed: the seed to use for splitting the dataset into train and test sets.
            lang: the spaCy language to use for training. If `None`, the language of the dataset will be used.
        """
        if isinstance(framework, str):
            framework = Framework(framework)

        # validate train and test sizes
        if train_size is None:
            train_size = 1
        if test_size is None:
            test_size = 1 - train_size

        # check if all numbers are larger than 0
        if not [abs(train_size), abs(test_size)] == [train_size, test_size]:
            raise ValueError("`train_size` and `test_size` must be larger than 0.")
        # check if train sizes sum up to 1
        if not (train_size + test_size) == 1:
            raise ValueError("`train_size` and `test_size` must sum to 1.")

        if test_size == 0:
            test_size = None

        if len(self.records) < 1:
            raise ValueError(
                "No records found in the dataset. Make sure you add records to the"
                " dataset via the `FeedbackDataset.add_records()` method first."
            )

        local_dataset = self
        if isinstance(task, (TrainingTaskForTextClassification, TrainingTaskForSentenceSimilarity)):
            if task.formatting_func is None:
                # in sentence-transformer models we can train without labels
                if task.label:
                    local_dataset = local_dataset.compute_unified_responses(
                        question=task.label.question, strategy=task.label.strategy
                    )
        elif isinstance(task, TrainingTaskForQuestionAnswering):
            if task.formatting_func is None:
                local_dataset = self.compute_unified_responses(question=task.answer.name, strategy="disagreement")
        elif not isinstance(
            task,
            (
                TrainingTaskForSFT,
                TrainingTaskForRM,
                TrainingTaskForPPO,
                TrainingTaskForDPO,
                TrainingTaskForChatCompletion,
            ),
        ):
            raise ValueError(f"Training data {type(task)} is not supported yet")

        return task.prepare_for_training(framework=framework, dataset=self, train_size=train_size, seed=seed, lang=lang)

    def update_records(self, records: Union["FeedbackRecord", List["FeedbackRecord"]]) -> None:
        warnings.warn(
            "`update_records` method only works for `FeedbackDataset` pushed to Argilla. "
            "If your are working with local data, you can just iterate over the records and update them."
        )

    def sort_by(
        self, field: Union[str, RecordSortField], order: Union[str, SortOrder] = SortOrder.asc
    ) -> "FeedbackDataset":
        warnings.warn(
            "`sort_by` method is not supported for local datasets and won't take any effect. "
            "First, you need to push the dataset to Argilla with `FeedbackDataset.push_to_argilla()`. "
            "After, use `FeedbackDataset.from_argilla(...).sort_by()`.",
            UserWarning,
            stacklevel=1,
        )
        return self

    def pull(self, *args, **kwargs) -> "FeedbackDataset":
        warnings.warn(
            "`pull` method is not supported for local datasets and won't take any effect."
            "First, you need to push the dataset to Argilla with `FeedbackDataset.push_to_argilla()`. "
            "After, use `FeedbackDataset.from_argilla(...).pull()`.",
            UserWarning,
        )
        return self

    def filter_by(self, *args, **kwargs) -> "FeedbackDataset":
        warnings.warn(
            "`filter_by` method is not supported for local datasets and won't take any effect. "
            "First, you need to push the dataset to Argilla with `FeedbackDataset.push_to_argilla()`. "
            "After, use `FeedbackDataset.from_argilla(...).filter_by()`.",
            UserWarning,
        )
        return self

    def delete(self):
        warnings.warn(
            "`delete` method is not supported for local datasets and won't take any effect. "
            "First, you need to push the dataset to Argilla with `FeedbackDataset.push_to_argilla`. "
            "After, use `FeedbackDataset.from_argilla(...).delete()`",
            UserWarning,
        )
        return self

    def find_similar_records(
        self,
        vector_name: str,
        value: Optional[List[float]] = None,
        record: Optional[R] = None,
        max_results: int = 50,
    ) -> List[Tuple[FeedbackRecord, float]]:
        warnings.warn(
            "`find_similar_records` method is not supported for local datasets and won't take any effect. "
            "First, you need to push the dataset to Argilla with `FeedbackDataset.push_to_argilla`. "
            "After, use `FeedbackDataset.from_argilla(...).find_similar_records()`",
            UserWarning,
        )
        return []



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path
from typing import Dict, Optional

from argilla.client.sdk.commons.errors import HttpResponseError, UnauthorizedApiError
from argilla.client.singleton import init
from argilla.pydantic_v1 import AnyHttpUrl, BaseModel

ARGILLA_CACHE_DIR = Path.home() / ".cache" / "argilla"
ARGILLA_CREDENTIALS_FILE = ARGILLA_CACHE_DIR / "credentials.json"


class ArgillaCredentials(BaseModel):
    api_url: AnyHttpUrl
    api_key: str
    workspace: Optional[str] = None
    extra_headers: Optional[Dict[str, str]] = None

    def save(self) -> None:
        with open(ARGILLA_CREDENTIALS_FILE, "w") as f:
            f.write(self.json())

    @classmethod
    def load(cls) -> "ArgillaCredentials":
        if not cls.exists():
            raise FileNotFoundError(f"'{ARGILLA_CREDENTIALS_FILE}' does not exist.")

        with open(ARGILLA_CREDENTIALS_FILE, "r") as f:
            return cls.parse_raw(f.read())

    @classmethod
    def remove(cls) -> None:
        if not cls.exists():
            raise FileNotFoundError(f"'{ARGILLA_CREDENTIALS_FILE}' does not exist.")

        ARGILLA_CREDENTIALS_FILE.unlink()

    @classmethod
    def exists(cls) -> bool:
        return ARGILLA_CREDENTIALS_FILE.exists()


def login(
    api_url: str, api_key: str, workspace: Optional[str] = None, extra_headers: Optional[Dict[str, str]] = None
) -> None:
    """Login to an Argilla server using the provided URL and API key. If the login is successful, the credentials will
    be stored in the Argilla cache directory (`~/.cache/argilla/credentials.json`).

    Args:
        api_url: The URL of the Argilla server.
        api_key: The API key to use when communicating with the Argilla server.
        workspace: The default workspace where the datasets will be created.
        extra_headers: A dictionary containing extra headers that will be sent to the Argilla server.

    Raises:
        ValueError: If the login fails.
    """
    # Try to login to the server
    try:
        init(api_url=api_url, api_key=api_key, workspace=workspace, extra_headers=extra_headers)
    except HttpResponseError as e:
        raise ValueError(
            f"Could not reach '{api_url}', make sure that the Argilla Server is running and working as expected"
        ) from e
    except UnauthorizedApiError as e:
        raise ValueError(f"Could not login in '{api_url}' using provided credentials") from e

    if not ARGILLA_CACHE_DIR.exists():
        ARGILLA_CACHE_DIR.mkdir(parents=True)

    ArgillaCredentials(api_url=api_url, api_key=api_key, workspace=workspace, extra_headers=extra_headers).save()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Optional
from uuid import UUID

import httpx

from argilla.pydantic_v1 import BaseModel, Field


class RemoteSchema(BaseModel, ABC):
    # TODO(@alvarobartt): Review optional id configuration for remote schemas
    id: Optional[UUID] = Field(None, allow_mutation=False)
    client: Optional[httpx.Client] = Field(None, allow_mutation=False)

    # TODO(@alvarobartt): here to be able to use the `allowed_for_roles` decorator
    @property
    def _client(self) -> Optional[httpx.Client]:
        return self.client

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    @abstractmethod
    def to_local(self) -> BaseModel:
        """Abstract method to be implemented by subclasses to convert the remote schema
        to a local one.
        """
        ...

    @classmethod
    @abstractmethod
    def from_api(cls, payload: BaseModel) -> "RemoteSchema":
        """Abstract method to be implemented by subclasses to convert the API payload
        into a remote schema."""
        ...



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""This module contains metrics to gather information related to inter-Annotator agreement. """

from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Union

from nltk.metrics.agreement import AnnotationTask as NLTKAnnotationTask
from nltk.metrics.distance import binary_distance, interval_distance, masi_distance

from argilla.client.feedback.dataset import FeedbackDataset
from argilla.client.feedback.metrics.base import AgreementMetricResult, AnnotationTaskMetricBase, MetricBase
from argilla.client.feedback.schemas import (
    LabelQuestion,
    MultiLabelQuestion,
    RankingQuestion,
    RatingQuestion,
)
from argilla.client.feedback.schemas.remote.shared import RemoteSchema

if TYPE_CHECKING:
    from argilla.client.feedback.dataset import FeedbackDataset
    from argilla.client.feedback.dataset.remote.dataset import RemoteFeedbackDataset
    from argilla.client.feedback.metrics.base import FormattedResponses
    from argilla.client.feedback.schemas.enums import ResponseStatusFilter
    from argilla.client.feedback.schemas.records import SortBy


def prepare_dataset_for_annotation_task(
    dataset: Union["FeedbackDataset", "RemoteFeedbackDataset"],
    question_name: str,
    filter_by: Optional[Dict[str, Union["ResponseStatusFilter", List["ResponseStatusFilter"]]]] = None,
    sort_by: Optional[List["SortBy"]] = None,
    max_records: Optional[int] = None,
) -> "FormattedResponses":
    """Helper function to prepare the dataset for the nltk's AnnotationTask.

    The AnnotationTask class from nltk expects the data to be formatted as a list
    of tuples, each containing the annotator id, the task id and the label.

    The AnnotationTask is supposed to deal with sets and hashable objects, but
    there are errors transforming the data to sets and using the MASI distance function.
    For the moment what we do with those type of questions is create a string
    with all the values, and use the edit distance function.

    Note:
        We could potentially extend the functionality to a more than a question name. The
        requirement would be that all the questions are of the same type, as that would
        determine the type of distance function to use.

    Args:
        dataset: FeedbackDataset to compute the metrics.
        question_name: Name of the question for which we want to analyse the agreement.
        filter_by: A dict with key the field to filter by, and values the filters to apply.
            Can be one of: draft, pending, submitted, and discarded. If set to None,
            no filter will be applied. Defaults to None (no filter is applied).
        sort_by: A list of `SortBy` objects to sort your dataset by.
            Defaults to None (no filter is applied).
        max_records: The maximum number of records to use for training. Defaults to None.

    Returns:
        formatted_responses: The responses formatted as a list of tuples of (user_id, question_id, value).
    """
    question_type = type(dataset.question_by_name(question_name))
    # Check to assume the remote questions behave just like local ones
    if issubclass(question_type, RemoteSchema):
        question_type = type(dataset.question_by_name(question_name).to_local())

    supported_question_types = list(QUESTION_TO_DISTANCE.keys())
    if question_type not in supported_question_types:
        raise NotImplementedError(
            f"Question '{question_name}' is of type '{question_type}', the supported question types are: {supported_question_types}."
        )

    if filter_by:
        dataset = dataset.filter_by(**filter_by)
    if sort_by:
        dataset = dataset.sort_by(sort_by)
    if max_records:
        dataset = dataset.pull(max_records=max_records)

    hf_dataset = dataset.format_as("datasets")

    formatted_responses: FormattedResponses = []

    for row in hf_dataset:
        responses_ = row[question_name]
        question_text = row["text"]
        for response in responses_:
            user_id = response["user_id"]
            if user_id is None:
                raise ValueError(
                    "Please push your dataset to argilla to have the user_id necessary for this computation."
                )

            value = response["value"]
            if value is None:
                continue
            # To avoid errors with the MASI distance function
            if isinstance(value, list):
                if len(value) == 0:
                    continue
            if question_type == RankingQuestion:
                value = tuple(value["rank"])
            elif question_type == MultiLabelQuestion:
                value = frozenset(value)

            formatted_responses.append((user_id, question_text, value))

    return formatted_responses


def kendall_tau_dist(x: List[int], y: List[int]) -> float:
    r"""Kendall tau distance.

    https://en.wikipedia.org/wiki/Kendall_tau_distance

    Args:
        x: Values of the first annotation.
        y: Values of the first annotation.

    Returns:
        distance: Kendall tau distance.

    Example:
        >>> import itertools
        >>> values = (1, 2, 3)
        >>> for i, a in enumerate(itertools.permutations(values, len(values))):
        ...     for j, b in enumerate(itertools.permutations(values, len(values))):
        ...             if j >= i:
        ...                     print((a, b), kendall_tau_dist(a,b))
        ...
        ((1, 2, 3), (1, 2, 3)) 0.0
        ((1, 2, 3), (1, 3, 2)) 0.3333333333333333
        ((1, 2, 3), (2, 1, 3)) 0.3333333333333333
        ((1, 2, 3), (2, 3, 1)) 0.6666666666666667
        ((1, 2, 3), (3, 1, 2)) 0.6666666666666667
        ((1, 2, 3), (3, 2, 1)) 1.0
        ...
    """
    from scipy.stats import kendalltau

    coef, _ = kendalltau(x, y)
    return 0.5 * (1 - coef)


QUESTION_TO_DISTANCE = {
    LabelQuestion: binary_distance,
    MultiLabelQuestion: masi_distance,
    RatingQuestion: interval_distance,
    RankingQuestion: kendall_tau_dist,
}


class AgreementMetric(MetricBase):
    """Main class to compute agreement metrics.

    Example:
        >>> import argilla as rg
        >>> from argilla.client.feedback.metrics import AgreementMetric
        >>> metric = AgreementMetric(dataset=dataset, question_name=question, filter_by={"response_status": "submitted"})
        >>> metrics_report = metric.compute("alpha")

    """

    def __init__(
        self,
        dataset: FeedbackDataset,
        question_name: str,
        filter_by: Optional[Dict[str, Union["ResponseStatusFilter", List["ResponseStatusFilter"]]]] = None,
        sort_by: Optional[List["SortBy"]] = None,
        max_records: Optional[int] = None,
    ) -> None:
        """Initialize a `AgreementMetric` object to compute agreement metrics.

        Args:
            dataset: FeedbackDataset to compute the metrics.
            question_name: Name of the question for which we want to analyse the agreement.
            filter_by: A dict with key the field to filter by, and values the filters to apply.
                Can be one of: draft, pending, submitted, and discarded. If set to None,
                no filter will be applied. Defaults to None (no filter is applied).
            sort_by: A list of `SortBy` objects to sort your dataset by.
                Defaults to None (no filter is applied).
            max_records: The maximum number of records to use for training. Defaults to None.
        """
        self._metrics_per_question = METRICS_PER_QUESTION
        super().__init__(dataset, question_name)
        self._filter_by = filter_by
        self._sort_by = sort_by
        self._max_records = max_records

    def compute(self, metric_names: Union[str, List[str]]) -> List[AgreementMetricResult]:
        """Computes the agreement metrics for the given question.

        Args:
            metric_names: name or list of names for the metrics to compute. i.e. `alpha`.
            kwargs: additional arguments to pass to the metric.

        Raises:
            ValueError: If the metric name is not supported for the given question.

        Returns:
            agreement_metrics: A list of `AgreementMetricResult` objects for the dataset.
        """
        metric_names = self._check_metrics(metric_names)
        metric_classes = self._get_metric_classes(metric_names)

        dataset = prepare_dataset_for_annotation_task(
            self._dataset,
            self._question_name,
            filter_by=self._filter_by,
            sort_by=self._sort_by,
            max_records=self._max_records,
        )

        distance_function = QUESTION_TO_DISTANCE[self._question_type]

        metrics = []
        for metric_name, metric_cls in metric_classes:
            metric = metric_cls(annotated_dataset=dataset, distance_function=distance_function)
            result = metric.compute()
            metrics.append(AgreementMetricResult(metric_name=metric_name, result=result, count=len(dataset)))

        if len(metric_names) == 1:
            return metrics[0]

        return metrics


class NLTKAnnotationTaskMetric(NLTKAnnotationTask, AnnotationTaskMetricBase):
    """Base class for metrics that use the nltk's AnnotationTask class.

    These metrics make use of a distance function to compute the distance between

    It is often the case that we don't want to treat two different
    labels as complete disagreement, and so the AnnotationTask constructor can also
    take a distance metric as a final argument. Distance metrics are functions that take two
    arguments, and return a value between 0.0 and 1.0 indicating the distance between them.

    By default, the following distance metrics are provided for each type of question:

        For LabelQuestion, binary_distance:

        >>> am.binary_distance("a", "b")
        1.0
        >>> am.binary_distance("a", "a")
        0.0

        For MultiLabelQuestion, masi_distance:

        >>> label_sets = [
        ...     [frozenset(["a", "b"]), frozenset(["b", "a"])],
        ...     [frozenset(["a"]), frozenset(["a", "b"])],
        ...     [frozenset(["c"]), frozenset(["a", "b"])],
        ... ]
        >>> for a, b in label_sets:
        ...     print((a,b), am.masi_distance(a,b))
        ...
        (frozenset({'a', 'b'}), frozenset({'a', 'b'})) 0.0
        (frozenset({'a'}), frozenset({'a', 'b'})) 0.665
        (frozenset({'c'}), frozenset({'a', 'b'})) 1.0

        For RatingQuestion, interval_distance:

        >>> for a, b in [(1, 1), (1, 2), (3,6)]:
        ...     print((a,b), am.interval_distance(a,b))
        ...
        (1, 1) 0
        (1, 2) 1
        (3, 6) 9

        For RankingQuestion, kendall_tau_dist:

        >>> for i, a in enumerate(itertools.permutations(values, len(values))):
        ...     for j, b in enumerate(itertools.permutations(values, len(values))):
        ...         if j >= i:
        ...             print((a, b), kendall_tau_dist(a,b))
        ...
        ((1, 2, 3), (1, 2, 3)) 0.0
        ((1, 2, 3), (1, 3, 2)) 0.3333333333333333
        ((1, 2, 3), (2, 1, 3)) 0.3333333333333333
        ((1, 2, 3), (2, 3, 1)) 0.6666666666666667
        ((1, 2, 3), (3, 1, 2)) 0.6666666666666667
        ((1, 2, 3), (3, 2, 1)) 1.0
        ((1, 3, 2), (1, 3, 2)) 0.0
        ...
    """

    def __init__(self, annotated_dataset: "FormattedResponses" = None, distance_function: Callable = None) -> None:
        AnnotationTaskMetricBase.__init__(
            self, annotated_dataset=annotated_dataset, distance_function=distance_function
        )
        super().__init__(data=annotated_dataset, distance=distance_function)


class KrippendorfAlpha(NLTKAnnotationTaskMetric):
    """Krippendorf's alpha agreement metric.

    Is a statistical measure of the inter-annotator agreement achieved when coding a set
    of units of analysis.

    To interpret the results from this metric, we refer the reader to the wikipedia entry.
    The common consensus dictates that a value of alpha >= 0.8 indicates a reliable annotation,
    a value >= 0.667 can only guarantee tentative conclusions, while lower values suggest an
    unreliable annotation.

    See Also:
        - Take a look at this metric definition:
        https://en.wikipedia.org/wiki/Krippendorff%27s_alpha

        - We use the implementation from nltk:
        https://www.nltk.org/api/nltk.metrics.agreement.html#nltk.metrics.agreement.AnnotationTask.alpha
    """

    def _compute(self, dataset) -> float:
        return self.alpha()


METRICS_PER_QUESTION = {
    LabelQuestion: {
        "alpha": KrippendorfAlpha,
    },
    MultiLabelQuestion: {
        "alpha": KrippendorfAlpha,
    },
    RatingQuestion: {
        "alpha": KrippendorfAlpha,
    },
    RankingQuestion: {
        "alpha": KrippendorfAlpha,
    },
}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections import defaultdict
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import numpy as np
from tqdm import tqdm

from argilla.client.feedback.dataset.remote.dataset import RemoteFeedbackDataset
from argilla.client.feedback.schemas import RankingQuestion, TextQuestion
from argilla.client.feedback.schemas.enums import ResponseStatusFilter
from argilla.client.feedback.schemas.remote.questions import RemoteRankingQuestion

if TYPE_CHECKING:
    from argilla.client.feedback.dataset import FeedbackDataset
    from argilla.client.feedback.metrics.base import Responses, Suggestions
    from argilla.client.feedback.schemas.enums import ResponseStatusFilter
    from argilla.client.feedback.schemas.records import SortBy


def get_responses_and_suggestions_per_user(
    dataset: Union["FeedbackDataset", "RemoteFeedbackDataset"],
    question_name: str,
    filter_by: Optional[Dict[str, Union["ResponseStatusFilter", List["ResponseStatusFilter"]]]] = None,
    sort_by: Optional[List["SortBy"]] = None,
    max_records: Optional[int] = None,
    show_progress: bool = True,
) -> Tuple[Dict[int, "Responses"], "Suggestions"]:
    """Extract the responses per user and the suggestions from a FeedbackDataset.

    Helper function for the metrics module where we want to compare the responses
    in relation to the suggestions offered.

    Args:
        dataset: FeedbackDataset or RemoteFeedbackDataset.
        question_name: The name of the question to filter from the dataset.
        filter_by: A dict with key the field to filter by, and values the filters to apply.
            Can be one of: draft, pending, submitted, and discarded. If set to None,
            no filter will be applied. Defaults to None (no filter is applied).
        sort_by: A list of `SortBy` objects to sort your dataset by.
            Defaults to None (no filter is applied).
        max_records: The maximum number of records to use for training. Defaults to None.

    Raises:
        NotImplementedError:
            When no user_id is given. We need that information to compute the metrics.

    Returns:
        Tuple containing the responses per user as a dict, with keys the user id and values the responses,
        and the suggestions.
    """
    if filter_by:
        dataset = dataset.filter_by(**filter_by)
    if sort_by:
        dataset = dataset.sort_by(sort_by)
    if max_records:
        dataset = dataset.pull(max_records=max_records)

    hf_dataset = dataset.format_as("datasets")
    question_type = type(dataset.question_by_name(question_name))
    is_ranking_question = (question_type == RankingQuestion) or (question_type == RemoteRankingQuestion)

    responses_and_suggestions_per_user = defaultdict(lambda: defaultdict(list))

    for responses_, suggestion in tqdm(
        zip(hf_dataset[question_name], hf_dataset[f"{question_name}-suggestion"]),
        desc="Extracting responses and suggestions per user",
        total=len(hf_dataset),
        disable=not show_progress,
    ):
        if is_ranking_question:
            suggestion = suggestion["rank"]

        if isinstance(suggestion, list):
            # To make it hashable
            suggestion = tuple(suggestion)

        for response in responses_:
            user_id = response["user_id"]
            if user_id is None:
                raise NotImplementedError(
                    "In order to use this functionality the records need to be assigned to a user."
                )

            if is_ranking_question:
                value = response["value"]["rank"]
            else:
                value = response["value"]

            if value is None:
                continue
            # To avoid errors with the MASI distance function
            if isinstance(value, list):
                if len(value) == 0:
                    continue

            if isinstance(value, list):
                # To make it hashable
                value = tuple(value)

            responses_and_suggestions_per_user[user_id]["responses"].append(value)
            responses_and_suggestions_per_user[user_id]["suggestions"].append(suggestion)

    return responses_and_suggestions_per_user


def get_unified_responses_and_suggestions(
    dataset: Union["FeedbackDataset", "RemoteFeedbackDataset"],
    question_name: str,
    strategy_name: str = "majority",
    filter_by: Optional[Dict[str, Union["ResponseStatusFilter", List["ResponseStatusFilter"]]]] = None,
    sort_by: Optional[List["SortBy"]] = None,
    max_records: Optional[int] = None,
) -> Tuple["Responses", "Suggestions"]:
    """Extract the unified responses and the suggestions from a FeedbackDataset.

    Helper function for the metrics module where we want to compare the responses
    in relation to the suggestions offered.

    Args:
        dataset: FeedbackDataset or RemoteFeedbackDataset.
        question_name: The name of the question to filter from the dataset.
        strategy_name: The name of the strategy to use to unify the responses.
        filter_by: A dict with key the field to filter by, and values the filters to apply.
            Can be one of: draft, pending, submitted, and discarded. If set to None,
            no filter will be applied. Defaults to None (no filter is applied).
        sort_by: A list of `SortBy` objects to sort your dataset by.
            Defaults to None (no filter is applied).
        max_records: The maximum number of records to use for training. Defaults to None.

    Raises:
        NotImplementedError:
            When asked for a TextQuestion.
        ValueError:
            If the dataset hasn't been unified yet.

    Returns:
        Tuple containing the unified responses and the suggestions.
    """
    if filter_by:
        dataset = dataset.filter_by(**filter_by)
    if sort_by:
        dataset = dataset.sort_by(sort_by)
    if max_records:
        dataset = dataset.pull(max_records=max_records)

    question_type = type(dataset.question_by_name(question_name))
    if question_type == TextQuestion:
        raise NotImplementedError("This function is not available for `TextQuestion`.")

    if isinstance(dataset, RemoteFeedbackDataset):
        dataset = dataset.pull()

    dataset.compute_unified_responses(question_name, strategy=strategy_name)

    unified_responses = []
    suggestions = []

    if not dataset.records[0].unified_responses:
        raise ValueError("Please unify the responses first: `dataset.compute_unified_responses(question, strategy)`.")

    # Get the position of the suggestion from the first record to avoid the nested loop,
    # and the type of the value to cast the unified responses.
    for idx_suggestion, suggestion in enumerate(dataset.records[0].suggestions):
        if suggestion.question_name == question_name:
            value_type = type(suggestion.value)
            break

    if value_type == list:
        # To make MultiLabelQuestion hashable
        value_type = tuple

    for record in dataset.records:
        unified_response = value_type(record.unified_responses[question_name][0].value)
        unified_responses.append(unified_response)
        suggestions.append(value_type(record.suggestions[idx_suggestion].value))

    if question_type == RankingQuestion:
        unified_responses = [
            tuple(ranking_schema.rank for ranking_schema in response) for response in unified_responses
        ]
        suggestions = [tuple(s["rank"] for s in suggestion) for suggestion in suggestions]

    return unified_responses, suggestions


def map_str_to_int(values: List[str]) -> List[int]:
    """Helper function to work with label questions as numerical values.

    Args:
        values: responses or suggestions with the string labels.

    Returns:
        values: corresponding values as integers to compute the metric
    """
    unique_values = np.unique(values)
    map_values = {value: i for i, value in enumerate(unique_values)}
    return [map_values[value] for value in values]


def is_multiclass(data) -> bool:
    """Helper function to check if a list of responses from LabelQuestion is also multiclass."""
    return len(np.unique(data)) > 2



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, Optional

import typer

from argilla.client.enums import DatasetType


def list_datasets(
    workspace: Optional[str] = typer.Option(None, help="Filter datasets by workspace"),
    type_: Optional[DatasetType] = typer.Option(
        None,
        "--type",
        help="The type of datasets to be listed. This option can be used multiple times. By default, all datasets are listed.",
    ),
) -> None:
    from rich.console import Console

    from argilla.cli.rich import echo_in_panel, get_argilla_themed_table
    from argilla.client.api import list_datasets as list_datasets_api
    from argilla.client.sdk.datasets.models import Dataset
    from argilla.client.workspaces import Workspace

    console = Console()

    def build_tags_text(tags: Dict[str, str]) -> str:
        text = ""
        for i, (tag, description) in enumerate(tags.items()):
            text += f"• [b]{tag}[not b]: {description}"
            if i < len(tags) - 1:
                text += "\n"
        return text

    table = get_argilla_themed_table(title="Datasets", show_lines=True)
    for column in ("ID", "Name", "Workspace", "Type", "Tags", "Creation Date", "Last Update Date"):
        table.add_column(column, justify="center" if column != "Tags" else "left")

    if workspace is not None:
        try:
            Workspace.from_name(workspace)
        except ValueError as e:
            echo_in_panel(
                f"Workspace with name={workspace} does not exist",
                title="Workspace not found",
                title_align="left",
                success=False,
            )
            raise typer.Exit(code=1) from e

    for dataset in list_datasets_api(workspace, type_):
        if isinstance(dataset, Dataset):
            table.add_row(
                dataset.id,
                dataset.name,
                dataset.workspace,
                dataset.task,
                build_tags_text(dataset.tags),
                dataset.created_at.isoformat(sep=" "),
                dataset.last_updated.isoformat(sep=" "),
            )
        else:
            table.add_row(
                str(dataset.id),
                dataset.name,
                dataset.workspace.name,
                "Feedback",
                None,
                dataset.created_at.isoformat(sep=" "),
                dataset.updated_at.isoformat(sep=" "),
            )

    console.print(table)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings

from argilla.cli import (
    datasets_app,
    info_app,
    login_app,
    logout_app,
    training_app,
    users_app,
    whoami_app,
    workspaces_app,
)
from argilla.cli.typer_ext import ArgillaTyper
from argilla.utils.dependency import is_package_with_extras_installed

warnings.simplefilter("ignore", UserWarning)

app = ArgillaTyper(help="Argilla CLI", no_args_is_help=True)


@app.error_handler(PermissionError)
def handler_permission_error(e: PermissionError) -> None:
    import sys

    from rich.console import Console

    from argilla.cli.rich import get_argilla_themed_panel

    panel = get_argilla_themed_panel(
        "Logged in user doesn't have enough permissions to execute this command",
        title="Not enough permissions",
        title_align="left",
        success=False,
    )

    Console().print(panel)
    sys.exit(1)


app.add_typer(datasets_app, name="datasets")
app.add_typer(info_app, name="info")
app.add_typer(login_app, name="login")
app.add_typer(logout_app, name="logout")
app.add_typer(training_app, name="train")
app.add_typer(users_app, name="users")
app.add_typer(whoami_app, name="whoami")
app.add_typer(workspaces_app, name="workspaces")

if is_package_with_extras_installed("argilla", ["server"]):
    from argilla_server.cli import app as server_app

    app.add_typer(server_app, name="server")

if __name__ == "__main__":
    app()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional

from argilla.client.feedback.schemas.validators import title_must_have_value
from argilla.pydantic_v1 import BaseModel, Field, PositiveInt, validator


class VectorSettings(BaseModel):
    """Schema for the `FeedbackDataset` vectors settings. The vectors setttings are used
    to define the configuration of the vectors associated to the records of a `FeedbackDataset`
    that will be used to perform the vector search.

    Args:
        name: The name of the vector settings.
        title: The title of the vector settings. If not provided, it will be capitalized
            from the `name` field. And its what will be shown in the UI. Defaults to
            `None`.
        dimensions: The dimensions of the vectors associated with the vector settings.

    Examples:
        >>> from argilla.client.feedback.schemas import VectorSettings
        >>> VectorSettings(name="my_vector_settings", dimensions=768)
    """

    name: str = Field(..., regex=r"^(?=.*[a-z0-9])[a-z0-9_-]+$")
    title: Optional[str] = None
    dimensions: PositiveInt

    _title_must_have_value = validator("title", always=True, allow_reuse=True)(title_must_have_value)
